#! /usr/bin/env python
# -*- coding: utf-8 -*-

from collections import defaultdict
import logging
import math
import os

from lab import tools

from downward.reports.plot import MatplotlibPlot, Matplotlib, PgfPlots, \
    PlotReport, MIN_AXIS


class ScatterMatplotlib(Matplotlib):
    @classmethod
    def _plot(cls, report, axes, categories, styles):
        # Display grid
        axes.grid(b=True, linestyle='-', color='0.75')

        has_points = False
        # Generate the scatter plots
        for category, coords in sorted(categories.items()):
            X, Y = zip(*coords)
            axes.scatter(X, Y, s=42, label=category, **styles[category])
            if X and Y:
                has_points = True

        if report.xscale == 'linear' or report.yscale == 'linear':
            # TODO: assert that both are linear or log
            plot_size = max(report.x_missing_val * 1.01, report.y_missing_val * 1.01)
        else:
            plot_size = max(report.x_missing_val * 1.5, report.y_missing_val * 1.5)

        # Plot a diagonal black line. Starting at (0,0) often raises errors.
        axes.plot([0.001, plot_size], [0.001, plot_size], 'k')

        axes.set_xlim(report.xlim_left or -1, report.xlim_right or plot_size)
        axes.set_ylim(report.ylim_bottom or -1, report.ylim_top or plot_size)
        # axes.set_xlim(report.xlim_left, report.xlim_right)
        # axes.set_ylim(report.ylim_bottom, report.ylim_top)

        for axis in [axes.xaxis, axes.yaxis]:
            # MatplotlibPlot.change_axis_formatter(
                # axis, report.missing_val if report.show_missing else None)
            MatplotlibPlot.change_axis_formatter(axes.xaxis,
                report.x_missing_val if report.show_missing else None)
            MatplotlibPlot.change_axis_formatter(axes.yaxis,
                report.y_missing_val if report.show_missing else None)
        return has_points


class ScatterPgfPlots(PgfPlots):
    @classmethod
    def _format_coord(cls, coord):
        def format_value(v):
            return str(v) if isinstance(v, int) else '%f' % v
        return '(%s, %s)' % (format_value(coord[0]), format_value(coord[1]))

    @classmethod
    def _get_plot(cls, report):
        lines = []
        options = cls._get_axis_options(report)
        lines.append('\\begin{axis}[%s]' % cls._format_options(options))
        for category, coords in sorted(report.categories.items()):
            plot = {'only marks': True}
            lines.append(
                '\\addplot+[%s] coordinates {\n%s\n};' % (
                    cls._format_options(plot),
                    ' '.join(cls._format_coord(c) for c in coords)))
            if category:
                lines.append('\\addlegendentry{%s}' % category)
            elif report.has_multiple_categories:
                # None is treated as the default category if using multiple
                # categories. Add a corresponding entry to the legend.
                lines.append('\\addlegendentry{default}')
        # Add black line.
        start = min(report.min_x, report.min_y)
        if report.xlim_left is not None:
            start = min(start, report.xlim_left)
        if report.ylim_bottom is not None:
            start = min(start, report.ylim_bottom)
        end = max(report.max_x, report.max_y)
        if report.xlim_right:
            end = max(end, report.xlim_right)
        if report.ylim_top:
            end = max(end, report.ylim_top)
        if report.show_missing:
            end = max(end, report.missing_val)
        lines.append(
            '\\addplot[color=black] coordinates {(%f, %f) (%d, %d)};' %
            (start, start, end, end))
        lines.append('\\end{axis}')
        return lines

    @classmethod
    def _get_axis_options(cls, report):
        opts = PgfPlots._get_axis_options(report)
        # Add line for missing values.
        for axis in ['x', 'y']:
            opts['extra %s ticks' % axis] = report.missing_val
            opts['extra %s tick style' % axis] = 'grid=major'
        return opts

class GeneralScatterPlotReport(PlotReport):
    """
    Generate a scatter plot for a specific attribute.
    """
    def __init__(self, x_algo, y_algo, x_attribute, y_attribute, show_missing=True, get_category=None, **kwargs):
        """
        See :class:`.PlotReport` for inherited arguments.

        The keyword argument *attributes* must contain exactly one
        attribute.

        Use the *filter_algorithm* keyword argument to select exactly
        two algorithms.

        If only one of the two algorithms has a value for a run, only
        add a coordinate if *show_missing* is True.

        *get_category* can be a function that takes **two** runs
        (dictionaries of properties) and returns a category name. This
        name is used to group the points in the plot. If there is more
        than one group, a legend is automatically added. Runs for which
        this function returns None are shown in a default category and
        are not contained in the legend. For example, to group by
        domain:

        >>> def domain_as_category(run1, run2):
        ...     # run2['domain'] has the same value, because we always
        ...     # compare two runs of the same problem.
        ...     return run1['domain']

        Example grouping by difficulty:

        >>> def improvement(run1, run2):
        ...     time1 = run1.get('search_time', 1800)
        ...     time2 = run2.get('search_time', 1800)
        ...     if time1 > time2:
        ...         return 'better'
        ...     if time1 == time2:
        ...         return 'equal'
        ...     return 'worse'

        >>> from downward.experiment import FastDownwardExperiment
        >>> exp = FastDownwardExperiment()
        >>> exp.add_report(ScatterPlotReport(
        ...     attributes=['search_time'],
        ...     get_category=improvement))

        Example comparing the number of expanded states for two
        algorithms:

        >>> exp.add_report(ScatterPlotReport(
        ...         attributes=["expansions_until_last_jump"],
        ...         filter_algorithm=["algorithm-1", "algorithm-2"],
        ...         get_category=domain_as_category,
        ...         format="png",  # Use "tex" for pgfplots output.
        ...         ),
        ...     name="scatterplot-expansions")

        """
        # If the size has not been set explicitly, make it a square.
        matplotlib_options = kwargs.get('matplotlib_options', {})
        matplotlib_options.setdefault('figure.figsize', [8, 8])
        kwargs['matplotlib_options'] = matplotlib_options
        PlotReport.__init__(self, **kwargs)
        if not self.attribute:
            logging.critical('ScatterPlotReport needs exactly one attribute')
        # By default all values are in the same category.
        self.get_category = get_category or (lambda run1, run2: None)
        self.show_missing = show_missing
        self.xlim_left = self.xlim_left or MIN_AXIS
        self.ylim_bottom = self.ylim_bottom or MIN_AXIS
        if self.output_format == 'tex':
            self.writer = ScatterPgfPlots
        else:
            self.writer = ScatterMatplotlib
        self.x_algo = x_algo
        self.y_algo = y_algo
        self.x_attribute = x_attribute
        self.y_attribute = y_attribute

    def _set_scales(self, xscale, yscale):
        PlotReport._set_scales(self, xscale or self.attribute.scale or 'log', yscale)
        if self.xscale != self.yscale:
            logging.critical('Scatterplots must use the same scale on both axes.')

    def _get_missing_val(self, max_value, scale):
        """
        Separate the missing values by plotting them at (max_value * 10)
        rounded to the next power of 10.
        """
        assert max_value is not None
        # HACK!
        max_value = 1800
        if scale == 'linear':
            return max_value * 1.1
        return int(10 ** math.ceil(math.log10(max_value)))

    def _handle_none_values(self, X, Y, replacement_x, replacement_y):
        assert len(X) == len(Y), (X, Y)
        if self.show_missing:
            return ([x if x is not None else replacement_x for x in X],
                    [y if y is not None else replacement_y for y in Y])
        return zip(*[(x, y) for x, y in zip(X, Y) if x is not None and y is not None])

    def _fill_categories(self, runs):
        # We discard the *runs* parameter.
        # Map category names to value tuples
        categories = defaultdict(list)
        x_count = 0
        y_count = 0
        x_none_count = 0
        y_none_count = 0
        for (domain, problem), runs in self.problem_runs.items():
            run1 = next((run for run in runs if run['algorithm'] == self.x_algo), None)
            run2 = next((run for run in runs if run['algorithm'] == self.y_algo), None)
            if run1 is None or run2 is None:
                continue
            assert (run1['algorithm'] == self.x_algo and
                    run2['algorithm'] == self.y_algo)
            val1 = run1.get(self.x_attribute)
            val2 = run2.get(self.y_attribute)
            x_count += 1
            y_count += 1
            if val1 is None:
                x_none_count += 1
            if val2 is None:
                y_none_count += 1
            # print val1, val2
            if val1 is None and val2 is None:
                continue
            category = self.get_category(run1, run2)
            categories[category].append((val1, val2))
        # print x_count, y_count
        # print x_none_count, y_none_count
        # print len(categories[None])
        # print categories[None]
        return categories

    def _get_limit(self, varlist, limit_type):
        assert limit_type == 'max' or limit_type == 'min'
        varlist = [x for x in varlist if x is not None]
        if(limit_type == 'max'):
            return max(varlist)
        else:
            return min(varlist)

    def _get_plot_size(self, missing_val, scale):
        if scale == 'linear':
            return missing_val * 1.01
        else:
            return missing_val * 1.25

    def _prepare_categories(self, categories):
        categories = PlotReport._prepare_categories(self, categories)

        # Find max-value to fit plot and to draw missing values.
        # self.missing_val = self._get_missing_val(max(self.max_x, self.max_y))
        self.x_missing_val = self._get_missing_val(self.max_x, self.xscale)
        self.y_missing_val = self._get_missing_val(self.max_y, self.yscale)
        # print self.x_missing_val, self.y_missing_val

        # set minima
        self.xlim_left = self._get_limit([self.xlim_left, self.min_x],'min')
        self.ylim_bottom = self._get_limit([self.ylim_bottom, self.min_y],'min')

        # set maxima
        x_plot_size = y_plot_size = None
        if self.show_missing:
            x_plot_size = self._get_plot_size(self.x_missing_val, self.xscale)
            y_plot_size = self._get_plot_size(self.y_missing_val, self.yscale)
        self.xlim_right = self._get_limit([self.xlim_right, self.max_x, x_plot_size], 'max')
        self.ylim_top = self._get_limit([self.ylim_top, self.max_y, y_plot_size], 'max')

        # self.diagonal_start = self.diagonal_end = None
        # if self.show_diagonal:
            # self.diagonal_start = max(self.xlim_left, self.ylim_bottom)
            # self.diagonal_end = min(self.xlim_right, self.ylim_top)

        new_categories = {}
        for category, coords in categories.items():
            X, Y = zip(*coords)
            # X, Y = self._handle_none_values(X, Y, self.missing_val)
            X, Y = self._handle_none_values(X, Y, self.x_missing_val, self.y_missing_val)
            coords = zip(X, Y)
            new_categories[category] = coords
        # print len(new_categories[None])
        # print new_categories[None]
        return new_categories

    def write(self):
        if not (len(self.algorithms) == 1 and self.x_algo == self.algorithms[0] and self.y_algo == self.algorithms[0]):
            logging.critical(
                'Scatter plots need exactly 1 algorithm that must match x_algo and y_algo: %s, %s, %s' % (self.algorithms, self.x_algo, self.y_algo))
        self.xlabel = self.xlabel or self.x_algo + ": " + self.x_attribute
        self.ylabel = self.ylabel or self.y_algo + ": " + self.y_attribute

        suffix = '.' + self.output_format
        if not self.outfile.endswith(suffix):
            self.outfile += suffix
        tools.makedirs(os.path.dirname(self.outfile))
        self._write_plot(self.runs.values(), self.outfile)



#! /usr/bin/env python2.6
# -*- coding: utf-8 -*-

import sys


TIMEOUT = 1800
EPSILON = 0.0001 # Don't care about this little time.


class Results(object):
    def __init__(self, file):
        self.configs, self.problems, self.data = self._parse_entries(file)
        self._add_missing_entries()
        assert len(self.data) == len(self.configs) * len(self.problems)
        self._discard_costs()

    def _parse_entries(self, file):
        configs = set()
        problems = set()
        data = {}
        best_costs = {}
        for entry_ in self._parse(file):
            entry = self._massage_entry(entry_)

            config = entry["config"]
            problem = entry["problem"]
            cost = entry["cost"]

            if "_wa" in config or "alt" in config or "fdss" in config:
                continue
            if "08" not in entry_["domain"]:
                if config.endswith("_0"):
                    config = config[:-2] + "_1"
            else:
                if config.endswith(("_0", "_2")):
                    continue

#            if not (entry["domain"].endswith("08'") or
#                entry["domain"].endswith("08-strips'")):
#                continue
#            if (entry["domain"].endswith("08'") or
#                entry["domain"].endswith("08-strips'")):
#                continue

            configs.add(config)
            problems.add(problem)
            if cost is not None:
                if problem not in best_costs:
                    best_costs[problem] = entry["cost"]
                else:
                    best_costs[problem] = min(best_costs[problem], cost)
            data[config, problem] = (entry["time"], entry["cost"])

        for (config, problem), (time, cost) in data.iteritems():
            ipc_score = None
            if time is not None:
                best_cost = best_costs[problem]
                assert cost is not None
                assert best_cost <= cost
                ipc_score = float(best_cost)/cost
            data[config, problem] = (time, cost, ipc_score)
        return sorted(configs), sorted(problems), data

    def _parse(self, file):
        properties = ["config", "domain", "problem",
                      "status", "total_time", "cost"]
        data = {}
        for line in file:
            left, match, right = line.partition("=")
            left = left.strip()
            right = right.strip()
            if not match:
                assert line.startswith("[")
                if not line.startswith("[["):
                    if data:
                        yield data
                    data = {}
            elif left in properties:
                assert left not in data
                data[left] = right

    def _massage_entry(self, entry):
        assert "config" in entry, entry
        assert "domain" in entry, entry
        assert "problem" in entry, entry
        assert "status" in entry, entry
        config = entry["config"].strip("'")
        if config.startswith("WORK-"):
            config = config[5:]
        domain = entry["domain"].strip("'")
        problem = entry["problem"].strip("'")
        status = entry["status"]
        assert status in ["'ok'", "'unsolved'"], entry

        if "08" in domain and "0000" in config:
            # M&S config with no action cost support
            # print "IGNORED:", entry
            time = None
            cost = None
        elif status == "'ok'":
            assert "total_time" in entry, entry
            assert "cost" in entry, entry
            time = float(entry["total_time"])
            assert time <= TIMEOUT + EPSILON, entry
            cost = int(entry["cost"])
        else:
            assert "cost" not in entry, entry
            time = None
            cost = None
        return dict(
            config=config,
            problem="%s:%s" % (domain, problem),
            time=time,
            cost=cost)

    def _add_missing_entries(self):
        for config in self.configs:
            for problem in self.problems:
                if (config, problem) not in self.data:
                    print "MISSING:", config, problem
                    self.data[config, problem] = (None, None, None)

    def _discard_costs(self):
        for problem in self.problems:
            for config in self.configs:
                time = self.data[config, problem][0]
                ipc_score = self.data[config, problem][2]
                self.data[config, problem] = (time, ipc_score)

    def solution_times(self, problem):
        result = {}
        for config in self.configs:
            time = self.data[config, problem][0]
            if time is not None:
                result[config] = time
        return result

    def solution_scores(self, problem):
        result = {}
        for config in self.configs:
            ipc_score = self.data[config, problem][1]
            if ipc_score is not None:
                result[config] = ipc_score
        return result

    def _total_solved(self):
        return sum(1 for problem in self.problems
                   if self.solution_times(problem))

    def dump_statistics(self):
        print len(self.data), "results"
        print len(self.configs), "configs"
        print len(self.problems), "problems"
        print self._total_solved(), "problems solved by someone"
        for config in sorted(self.configs):
            num_solved = len([problem for problem in self.problems
                              if self.data[config, problem][0] is not None])
            print num_solved, "problems solved by", config
        for config in sorted(self.configs):
            ipc_score = sum([self.data[config, problem][1]
                             for problem in self.problems
                             if self.data[config, problem][1] is not None])
            print "%.2f" % ipc_score, "IPC score by", config


class Portfolio(object):
    def __init__(self, results, timeouts):
        self.results = results
        self.timeouts = timeouts
        self.configs = sorted(self.timeouts.iterkeys())

    def solves_problem(self, problem):
        solution_times = self.results.solution_times(problem)
        for solver, time in solution_times.iteritems():
            if time < self.timeouts[solver] + EPSILON:
                return True
        return False

    def total_timeout(self):
        return sum(self.timeouts.itervalues())

    def num_solved(self):
        return sum(1 for problem in self.results.problems
                   if self.solves_problem(problem))

    def score(self, problem):
        solution_times = self.results.solution_times(problem)
        solution_scores = self.results.solution_scores(problem)
        best_score = 0.0
        for solver, time in solution_times.iteritems():
            if time < self.timeouts[solver] + EPSILON:
                if solution_scores[solver] > best_score:
                    best_score = solution_scores[solver]
        return best_score

    def ipc_score(self):
        return sum(self.score(problem)
                   for problem in self.results.problems)

    def successors(self, granularity):
        for config in self.configs:
            succ_timeouts = self.timeouts.copy()
            succ_timeouts[config] += granularity
            yield Portfolio(self.results, succ_timeouts)

    def reduce(self, granularity):
        num_solved = self.num_solved()
        for config in self.configs:
            # Reduce timeout for this config until we hit zero or lose
            # a problem.
            while self.timeouts[config] > granularity + EPSILON:
                self.timeouts[config] -= granularity
                if self.num_solved() < num_solved:
                    # Undo last change and stop reducing this timeout.
                    self.timeouts[config] += granularity
                    break

    def reduce_score_based(self, granularity):
        score = self.ipc_score()
        for config in self.configs:
            # Reduce timeout for this config until we hit zero or lose
            # a problem.
            while self.timeouts[config] > granularity + EPSILON:
                self.timeouts[config] -= granularity
                if self.ipc_score() < score:
                    # Undo last change and stop reducing this timeout.
                    self.timeouts[config] += granularity
                    break

    def dump_marginal_contributions(self):
        print "Marginal contributions:"
        num_solved = self.num_solved()
        score = self.ipc_score()
        for config in self.configs:
            if self.timeouts[config] > EPSILON:
                succ_timeouts = self.timeouts.copy()
                succ_timeouts[config] = 0
                succ = Portfolio(self.results, succ_timeouts)
                num_lost = num_solved - succ.num_solved()
                score_lost = score - succ.ipc_score()
                print "   %3d problems / %.2f score from %s" % (num_lost,
                        score_lost, config)

    def dump_detailed_marginal_contributions(self, granularity):
        print "Detailed marginal contributions:"
        num_solved = self.num_solved()
        for config in self.configs:
            if self.timeouts[config] > EPSILON:
                reduction = 0
                prev_num_solved = num_solved
                lost_at = []
                while reduction <= self.timeouts[config] + EPSILON:
                    reduced_timeout = self.timeouts[config] - reduction
                    succ_timeouts = self.timeouts.copy()
                    succ_timeouts[config] = reduced_timeout
                    succ = Portfolio(self.results, succ_timeouts)
                    succ_solved = succ.num_solved()
                    if succ_solved < prev_num_solved:
                        for _ in xrange(prev_num_solved - succ_solved):
                            lost_at.append(str(reduction))
                        prev_num_solved = succ_solved
                    reduction += granularity
                print "   %s: problems lost at %s" % (
                    config, ", ".join(lost_at))

    def dump(self):
        print "portfolio for %.2f seconds solves %d problems with score %.2f:" % (
            self.total_timeout(), self.num_solved(), self.ipc_score())
        for config in self.configs:
            print "   %7.2f seconds for %s" % (self.timeouts[config], config)

    def dump_unsolved(self):
        print "Unsolved problems:"
        count = 0
        for problem in self.results.problems:
            solution_times = self.results.solution_times(problem)
            if solution_times and not self.solves_problem(problem):
                formatted_solution_times = "{%s}" % ", ".join(
                    "%s: %.2f" % pair
                    for pair in sorted(solution_times.items()))
                print "   %-35s %s" % (problem, formatted_solution_times)
                count += 1
        print "(%d problems)" % count


def compute_portfolio(results, granularity):
    portfolio = Portfolio(results, dict.fromkeys(results.configs, 0))
    while True:
        portfolio.dump()
        print
        best_successor = None
        best_solved = -1
        for successor in portfolio.successors(granularity):
            succ_solved = successor.num_solved()
            if succ_solved > best_solved:
                best_solved = succ_solved
                best_successor = successor
        if best_successor.total_timeout() > TIMEOUT + EPSILON:
            return portfolio
        portfolio = best_successor

def compute_portfolio_using_ipc_scores(results, granularity):
    portfolio = Portfolio(results, dict.fromkeys(results.configs, 0))
    while True:
        portfolio.dump()
        print
        best_successor = None
        best_score = -1
        for successor in portfolio.successors(granularity):
            succ_score = successor.ipc_score()
            if succ_score > best_score:
                best_score = succ_score
                best_successor = successor
        if best_successor.total_timeout() > TIMEOUT + EPSILON:
            return portfolio
        portfolio = best_successor

def main():
    if len(sys.argv) >= 3:
        granularity = int(sys.argv[2])
    results = Results(open(sys.argv[1]))
    results.dump_statistics()
#    print "Computing portfolio using coverage..."
#    portfolio = compute_portfolio(results, granularity=120)
#    print
#    portfolio.dump()
#    # pprint.pprint(results.data)
#    print
#    print "Reducing portfolio..."
#    portfolio.reduce(granularity=1)
#    portfolio.dump()
#    print
#    portfolio.dump_unsolved()
#    print
#    portfolio.dump_marginal_contributions()
#    print
#    portfolio.dump_detailed_marginal_contributions(granularity=1)
#    print
#    print "Computing portfolio using IPC scores..."
    portfolio = compute_portfolio_using_ipc_scores(results, granularity=granularity)
    print
    portfolio.dump()
    # pprint.pprint(results.data)
    print
    print "Reducing portfolio..."
    portfolio.reduce_score_based(granularity=1)
    portfolio.dump()
    print
    portfolio.dump_unsolved()
    print
    portfolio.dump_marginal_contributions()
    print
    portfolio.dump_detailed_marginal_contributions(granularity=1)
    print


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import os.path

from lab.environments import MaiaEnvironment
from lab.steps import Step

from downward.checkouts import Translator, Preprocessor, Planner
from downward.experiments import DownwardExperiment
from downward.reports.compare import CompareRevisionsReport
from downward.reports.scatter import ScatterPlotReport


def get_script():
    """Get file name of main script."""
    import __main__
    return __main__.__file__


def get_script_dir():
    """Get directory of main script.

    Usually a relative directory (depends on how it was called by the user.)"""
    return os.path.dirname(get_script())


def get_experiment_name():
    """Get name for experiment.

    Derived from the filename of the main script, e.g.
    "/ham/spam/eggs.py" => "eggs"."""
    script = os.path.abspath(get_script())
    script_dir = os.path.basename(os.path.dirname(script))
    script_base = os.path.splitext(os.path.basename(script))[0]
    return "%s-%s" % (script_dir, script_base)


def get_data_dir():
    """Get data dir for the experiment.

    This is the subdirectory "data" of the directory containing
    the main script."""
    return os.path.join(get_script_dir(), "data", get_experiment_name())


def get_repo_base():
    """Get base directory of the repository, as an absolute path.

    Found by searching upwards in the directory tree from the main
    script until a directory with a subdirectory named ".hg" is found."""
    path = os.path.abspath(get_script_dir())
    while True:
        if os.path.exists(os.path.join(path, ".hg")):
            return path
        path = os.path.dirname(path)


class MyExperiment(DownwardExperiment):
    DEFAULT_TABLE_ATTRIBUTES = [
        "cost",
        "coverage",
        "evaluations",
        "expansions",
        "expansions_until_last_jump",
        "generated",
        "memory",
        "run_dir",
        "score_evaluations",
        "score_expansions",
        "score_generated",
        "score_memory",
        "score_search_time",
        "score_total_time",
        "search_time",
        "total_time",
        ]

    DEFAULT_SCATTER_PLOT_ATTRIBUTES = [
        "total_time",
        "search_time",
        "memory",
        "expansions_until_last_jump",
        ]

    """Wrapper for DownwardExperiment with a few convenience features."""

    def __init__(self, configs=None, grid_priority=None, path=None,
                 repo=None, revisions=None, search_revisions=None,
                 suite=None, **kwargs):
        """Create a DownwardExperiment with some convenience features.

        If "configs" is specified, it should be a dict of {nick:
        cmdline} pairs that sets the planner configurations to test.

        If "grid_priority" is specified and no environment is
        specifically requested in **kwargs, use the maia environment
        with the specified priority.

        If "path" is not specified, the experiment data path is
        derived automatically from the main script's filename.

        If "repo" is not specified, the repository base is derived
        automatically from the main script's path.

        If "revisions" is specified, it should be a non-empty
        list of revisions, which specify which planner versions to use
        in the experiment. The same versions are used for translator,
        preprocessor and search.

        If "search_revisions" is specified, it should be a non-empty
        list of revisions, which specify which search component
        versions to use in the experiment. All experiments use the
        translator and preprocessor component of the first
        revision.

        If "suite" is specified, it should specify a problem suite.

        Options "combinations" (from the base class), "revisions" and
        "search_revisions" are mutually exclusive."""

        if grid_priority is not None and "environment" not in kwargs:
            kwargs["environment"] = MaiaEnvironment(priority=grid_priority)

        if path is None:
            path = get_data_dir()

        if repo is None:
            repo = get_repo_base()

        num_rev_opts_specified = (
            int(revisions is not None) +
            int(search_revisions is not None) +
            int(kwargs.get("combinations") is not None))

        if num_rev_opts_specified > 1:
            raise ValueError('must specify exactly one of "revisions", '
                             '"search_revisions" or "combinations"')

        # See add_comparison_table_step for more on this variable.
        self._HACK_revisions = revisions

        if revisions is not None:
            if not revisions:
                raise ValueError("revisions cannot be empty")
            combinations = [(Translator(repo, rev),
                             Preprocessor(repo, rev),
                             Planner(repo, rev))
                            for rev in revisions]
            kwargs["combinations"] = combinations

        if search_revisions is not None:
            if not search_revisions:
                raise ValueError("search_revisions cannot be empty")
            base_rev = search_revisions[0]
            translator = Translator(repo, base_rev)
            preprocessor = Preprocessor(repo, base_rev)
            combinations = [(translator, preprocessor, Planner(repo, rev))
                            for rev in search_revisions]
            kwargs["combinations"] = combinations

        DownwardExperiment.__init__(self, path=path, repo=repo, **kwargs)

        if configs is not None:
            for nick, config in configs.items():
                self.add_config(nick, config)

        if suite is not None:
            self.add_suite(suite)

        self._report_prefix = get_experiment_name()

    def add_comparison_table_step(self, attributes=None):
        revisions = self._HACK_revisions
        if revisions is None:
            # TODO: It's not clear to me what a "revision" in the
            # overall context of the code really is, e.g. when keeping
            # the translator and preprocessor method fixed and only
            # changing the search component. It's also not really
            # clear to me how the interface of the Compare... reports
            # works and how to use it more generally. Hence the
            # present hack.

            # Ideally, this method should look at the table columns we
            # have (defined by planners and planner configurations),
            # pair them up in a suitable way, either controlled by a
            # convenience parameter or a more general grouping method,
            # and then use this to define which pairs go together.
            raise NotImplementedError(
                "only supported when specifying revisions in __init__")

        if attributes is None:
            attributes = self.DEFAULT_TABLE_ATTRIBUTES
        if len(revisions) == 2:
            report = CompareRevisionsReport(*revisions, attributes=attributes)
            self.add_report(report, outfile="%s-compare.html" % self._report_prefix)
        else:
            # HACK: assumes the first revision is the noe everything else is compared against
            for rev in revisions[1:]:
                report = CompareRevisionsReport(revisions[0], rev, attributes=attributes)
                self.add_report(report, outfile="%s-compare-%s.html" % (self._report_prefix, rev))
            

    def add_scatter_plot_step(self, attributes=None):
        if attributes is None:
            attributes = self.DEFAULT_SCATTER_PLOT_ATTRIBUTES
        revisions = self._HACK_revisions
        if revisions is None:
            # TODO: See add_comparison_table_step.
            raise NotImplementedError(
                "only supported when specifying revisions in __init__")
        # HACK: assumes the first revision is the noe everything else is compared against
        for rev in revisions[1:]:
            # TODO: Should generalize this, too, by offering a general
            # grouping function and then comparing any pair of
            # settings in the same group.
            scatter_dir = os.path.join(self.eval_dir, "scatter")
            def make_scatter_plots():
                configs = [conf[0] for conf in self.configs]
                for nick in configs:
                    config_before = "%s-%s" % (revisions[0], nick)
                    config_after = "%s-%s" % (rev, nick)
                    for attribute in attributes:
                        name = "%s-%s-%s" % (self._report_prefix, attribute, nick)
                        report = ScatterPlotReport(
                            filter_config=[config_before, config_after],
                            attributes=[attribute],
                            get_category=lambda run1, run2: run1["domain"],
                            legend_location=(1.3, 0.5))
                        report(self.eval_dir, os.path.join(scatter_dir, rev, name))

            self.add_step(Step("make-scatter-plots", make_scatter_plots))



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import os

from lab.environments import LocalEnvironment, BaselSlurmEnvironment

import common_setup
from common_setup import IssueConfig, IssueExperiment


DIR = os.path.dirname(os.path.abspath(__file__))
BENCHMARKS_DIR = os.environ["DOWNWARD_BENCHMARKS"]
REVISIONS = ["issue731-base", "issue731-v4"]
BUILDS = ["release32"]
SEARCHES = [
    ("blind", "astar(blind())"),
    ("divpot", "astar(diverse_potentials())"),
    ("cegar", "astar(cegar())"),
    ("systematic2", "astar(cpdbs(systematic(2)))"),
    ("ipdb", "astar(ipdb())"),
]
CONFIGS = [
    IssueConfig(
        "{nick}-{build}".format(**locals()),
        ["--search", search],
        build_options=[build],
        driver_options=["--build", build])
    for nick, search in SEARCHES
    for build in BUILDS
]
SUITE = common_setup.DEFAULT_OPTIMAL_SUITE
ENVIRONMENT = BaselSlurmEnvironment(
    priority=0, email="jendrik.seipp@unibas.ch")

if common_setup.is_test_run():
    SUITE = IssueExperiment.DEFAULT_TEST_SUITE
    ENVIRONMENT = LocalEnvironment(processes=1)

exp = IssueExperiment(
    revisions=REVISIONS,
    configs=CONFIGS,
    environment=ENVIRONMENT,
)
exp.add_suite(BENCHMARKS_DIR, SUITE)
exp.add_comparison_table_step()

exp.run_steps()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Run some syntax checks. Return 0 if all tests pass and 1 otherwise.

The file bitbucket-pipelines.yml shows how to install the dependencies.
"""

from __future__ import print_function

import errno
import glob
import json
import os
import pipes
import re
import subprocess
import sys

DIR = os.path.dirname(os.path.abspath(__file__))
REPO = os.path.dirname(os.path.dirname(DIR))
SRC_DIR = os.path.join(REPO, "src")


def _get_src_files(path, extensions, ignore_dirs=None):
    ignore_dirs = ignore_dirs or []
    src_files = []
    for root, dirs, files in os.walk(path):
        for ignore_dir in ignore_dirs:
            if ignore_dir in dirs:
                dirs.remove(ignore_dir)
        src_files.extend([
            os.path.join(root, file)
            for file in files if file.endswith(extensions)])
    return src_files


def check_translator_style():
    output = subprocess.check_output([
        "./reindent.py", "--dryrun", "--recurse", "--verbose",
        os.path.join(SRC_DIR, "translate")], cwd=DIR).decode("utf-8")
    ok = True
    for line in output.splitlines():
        match = re.match("^checking (.+) ... changed.$", line)
        if match:
            ok = False
            print('Wrong format detected in %s. '
                  'Please run "./reindent.py -r ../src/translate"' %
                  match.group(1))
    return ok


def _run_pyflakes(path):
    python_files = _get_src_files(path, (".py",))
    python_files = [f for f in python_files if not f.endswith("__init__.py")]
    print("Checking {} files with pyflakes".format(len(python_files)))
    try:
        return subprocess.check_call(["pyflakes"] + python_files) == 0
    except OSError as err:
        if err.errno == 2:
            print(
                "Python style checks need pyflakes. Please install it "
                "with \"sudo apt-get install pyflakes\".")
        else:
            raise

def check_translator_pyflakes():
    return _run_pyflakes(os.path.join(SRC_DIR, "translate"))

def check_driver_pyflakes():
    return _run_pyflakes(os.path.join(REPO, "driver"))


def check_include_guard_convention():
    return subprocess.call("./check-include-guard-convention.py", cwd=DIR) == 0


def check_cc_files():
    """
    Currently, we only check that there is no "std::" in .cc files.
    """
    search_dir = os.path.join(SRC_DIR, "search")
    cc_files = _get_src_files(search_dir, (".cc",))
    print("Checking style of {} *.cc files".format(len(cc_files)))
    return subprocess.call(["./check-cc-file.py"] + cc_files, cwd=DIR) == 0


def check_cplusplus_style():
    """
    Calling the uncrustify mercurial extension doesn't work for
    bitbucket pipelines (the local .hg/hgrc file is untrusted and
    ignored by mercurial). We therefore find the source files manually
    and call uncrustify directly.
    """
    src_files = _get_src_files(REPO, (".h", ".cc"), ignore_dirs=["builds", "data"])
    print("Checking {} files with uncrustify".format(len(src_files)))
    config_file = os.path.join(REPO, ".uncrustify.cfg")
    executable = "uncrustify"
    try:
        returncode = subprocess.call(
            [executable, "-q", "-c", config_file, "--check"] + src_files,
        cwd=DIR, stdout=subprocess.PIPE)
    except OSError as err:
        if err.errno == errno.ENOENT:
            sys.exit("Error: {} not found. Is it on the PATH?".format(executable))
        else:
            raise
    if returncode != 0:
        print(
            'Run "hg uncrustify" to fix the coding style in the above '
            'mentioned files.')
    return returncode == 0


def check_search_code_with_clang_tidy():
    # clang-tidy needs the CMake files.
    build_dir = os.path.join(REPO, "builds", "clang-tidy")
    if not os.path.exists(build_dir):
        os.makedirs(build_dir)
    with open(os.devnull, 'w') as devnull:
        subprocess.check_call(["cmake", "../../src"], cwd=build_dir, stdout=devnull)

    # Create custom compilation database file. CMake outputs part of this information
    # when passing -DCMAKE_EXPORT_COMPILE_COMMANDS=ON, but the resulting file
    # contains no header files.
    search_dir = os.path.join(REPO, "src/search")
    src_files = _get_src_files(search_dir, (".h", ".cc"))
    compile_commands = [{
        "directory": os.path.join(build_dir, "search"),
        "command": "g++ -I{search_dir}/ext -std=c++11 -c {src_file}".format(**locals()),
        "file": src_file}
        for src_file in src_files
    ]
    with open(os.path.join(build_dir, "compile_commands.json"), "w") as f:
        json.dump(compile_commands, f, indent=2)

    # See https://clang.llvm.org/extra/clang-tidy/checks/list.html for
    # an explanation of the checks. We comment out inactive checks of some
    # categories instead of deleting them to see which additional checks
    # we could activate.
    checks = [
        # Enable with CheckTriviallyCopyableMove=0 when we require
        # clang-tidy >= 6.0 (see issue856).
        #"misc-move-const-arg",
        "misc-move-constructor-init",
        "misc-use-after-move",

        "performance-for-range-copy",
        "performance-implicit-cast-in-loop",
        "performance-inefficient-vector-operation",

        "readability-avoid-const-params-in-decls",
        #"readability-braces-around-statements",
        "readability-container-size-empty",
        "readability-delete-null-pointer",
        "readability-deleted-default",
        #"readability-else-after-return",
        #"readability-function-size",
        #"readability-identifier-naming",
        #"readability-implicit-bool-cast",
        # Disabled since we prefer a clean interface over consistent names.
        #"readability-inconsistent-declaration-parameter-name",
        "readability-misleading-indentation",
        "readability-misplaced-array-index",
        #"readability-named-parameter",
        #"readability-non-const-parameter",
        "readability-redundant-control-flow",
        "readability-redundant-declaration",
        "readability-redundant-function-ptr-dereference",
        "readability-redundant-member-init",
        "readability-redundant-smartptr-get",
        "readability-redundant-string-cstr",
        "readability-redundant-string-init",
        "readability-simplify-boolean-expr",
        "readability-static-definition-in-anonymous-namespace",
        "readability-uniqueptr-delete-release",
        ]
    cmd = [
        "./run-clang-tidy.py",
        "-quiet",
        "-p", build_dir,
        "-clang-tidy-binary=clang-tidy-5.0",
        # Include all non-system headers (.*) except the ones from search/ext/.
        "-header-filter=.*,-tree.hh,-tree_util.hh",
        "-checks=-*," + ",".join(checks)]
    print("Running clang-tidy: " + " ".join(pipes.quote(x) for x in cmd))
    print()
    try:
        output = subprocess.check_output(cmd, cwd=DIR, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError:
        print("Failed to run clang-tidy-5.0. Is it on the PATH?")
        return False
    errors = re.findall(r"^(.*:\d+:\d+: (?:warning|error): .*)$", output, flags=re.M)
    for error in errors:
        print(error)
    if errors:
        fix_cmd = cmd + [
            "-clang-apply-replacements-binary=clang-apply-replacements-5.0", "-fix"]
        print()
        print("You may be able to fix these issues with the following command: " +
            " ".join(pipes.quote(x) for x in fix_cmd))
    return not errors


def main():
    results = []
    for test_name, test in sorted(globals().items()):
        if test_name.startswith("check_"):
            print("Running {}".format(test_name))
            results.append(test())
    if all(results):
        print("All style checks passed")
    else:
        sys.exit("Style checks failed")


if __name__ == "__main__":
    main()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import os
import suites
from lab.reports import Attribute, gm

from common_setup import IssueConfig, IssueExperiment
try:
    from relativescatter import RelativeScatterPlotReport
    matplotlib = True
except ImportError:
    print 'matplotlib not availabe, scatter plots not available'
    matplotlib = False

def main(revisions=None):
    benchmarks_dir=os.path.expanduser('~/repos/downward/benchmarks')
    suite=suites.suite_optimal_strips()

    configs = {
        IssueConfig('rl-b50k', ['--search', 'astar(merge_and_shrink(merge_strategy=merge_precomputed(merge_tree=linear(variable_order=reverse_level)),shrink_strategy=shrink_bisimulation(greedy=false),label_reduction=exact(before_shrinking=true,before_merging=false),max_states=50000,threshold_before_merge=1))']),
        IssueConfig('cggl-b50k', ['--search', 'astar(merge_and_shrink(merge_strategy=merge_precomputed(merge_tree=linear(variable_order=cg_goal_level)),shrink_strategy=shrink_bisimulation(greedy=false),label_reduction=exact(before_shrinking=true,before_merging=false),max_states=50000,threshold_before_merge=1))']),
        IssueConfig('dfp-b50k', ['--search', 'astar(merge_and_shrink(merge_strategy=merge_stateless(merge_selector=score_based_filtering(scoring_functions=[goal_relevance,dfp,total_order])),shrink_strategy=shrink_bisimulation(greedy=false),label_reduction=exact(before_shrinking=true,before_merging=false),max_states=50000,threshold_before_merge=1))']),
        IssueConfig('rl-ginf', ['--search', 'astar(merge_and_shrink(merge_strategy=merge_precomputed(merge_tree=linear(variable_order=reverse_level)),shrink_strategy=shrink_bisimulation(greedy=true),label_reduction=exact(before_shrinking=true,before_merging=false),max_states=infinity,threshold_before_merge=1))']),
        IssueConfig('cggl-ginf', ['--search', 'astar(merge_and_shrink(merge_strategy=merge_precomputed(merge_tree=linear(variable_order=cg_goal_level)),shrink_strategy=shrink_bisimulation(greedy=true),label_reduction=exact(before_shrinking=true,before_merging=false),max_states=infinity,threshold_before_merge=1))']),
        IssueConfig('dfp-ginf', ['--search', 'astar(merge_and_shrink(merge_strategy=merge_stateless(merge_selector=score_based_filtering(scoring_functions=[goal_relevance,dfp,total_order])),shrink_strategy=shrink_bisimulation(greedy=true),label_reduction=exact(before_shrinking=true,before_merging=false),max_states=infinity,threshold_before_merge=1))']),
        IssueConfig('rl-f50k', ['--search', 'astar(merge_and_shrink(merge_strategy=merge_precomputed(merge_tree=linear(variable_order=reverse_level)),shrink_strategy=shrink_fh(),label_reduction=exact(before_shrinking=false,before_merging=true),max_states=50000))']),
        IssueConfig('cggl-f50k', ['--search', 'astar(merge_and_shrink(merge_strategy=merge_precomputed(merge_tree=linear(variable_order=cg_goal_level)),shrink_strategy=shrink_fh(),label_reduction=exact(before_shrinking=false,before_merging=true),max_states=50000))']),
        IssueConfig('dfp-f50k', ['--search', 'astar(merge_and_shrink(merge_strategy=merge_stateless(merge_selector=score_based_filtering(scoring_functions=[goal_relevance,dfp,total_order])),shrink_strategy=shrink_fh(),label_reduction=exact(before_shrinking=false,before_merging=true),max_states=50000))']),
    }

    exp = IssueExperiment(
        benchmarks_dir=benchmarks_dir,
        suite=suite,
        revisions=revisions,
        configs=configs,
        test_suite=['depot:p01.pddl'],
        processes=4,
        email='silvan.sievers@unibas.ch',
    )
    exp.add_resource('ms_parser', 'ms-parser.py', dest='ms-parser.py')
    exp.add_command('ms-parser', ['ms_parser'])

    # planner outcome attributes
    perfect_heuristic = Attribute('perfect_heuristic', absolute=True, min_wins=False)
    proved_unsolvability = Attribute('proved_unsolvability', absolute=True, min_wins=False)
    actual_search_time = Attribute('actual_search_time', absolute=False, min_wins=True, functions=[gm])

    # m&s attributes
    ms_construction_time = Attribute('ms_construction_time', absolute=False, min_wins=True, functions=[gm])
    ms_atomic_construction_time = Attribute('ms_atomic_construction_time', absolute=False, min_wins=True, functions=[gm])
    ms_abstraction_constructed = Attribute('ms_abstraction_constructed', absolute=True, min_wins=False)
    ms_final_size = Attribute('ms_final_size', absolute=False, min_wins=True)
    ms_out_of_memory = Attribute('ms_out_of_memory', absolute=True, min_wins=True)
    ms_out_of_time = Attribute('ms_out_of_time', absolute=True, min_wins=True)
    search_out_of_memory = Attribute('search_out_of_memory', absolute=True, min_wins=True)
    search_out_of_time = Attribute('search_out_of_time', absolute=True, min_wins=True)

    extra_attributes = [
        perfect_heuristic,
        proved_unsolvability,
        actual_search_time,

        ms_construction_time,
        ms_atomic_construction_time,
        ms_abstraction_constructed,
        ms_final_size,
        ms_out_of_memory,
        ms_out_of_time,
        search_out_of_memory,
        search_out_of_time,
    ]
    attributes = exp.DEFAULT_TABLE_ATTRIBUTES
    attributes.extend(extra_attributes)

    exp.add_comparison_table_step()

    #if matplotlib:
        #for attribute in ["memory", "total_time"]:
            #for config in configs:
                #exp.add_report(
                    #RelativeScatterPlotReport(
                        #attributes=[attribute],
                        #filter_config=["{}-{}".format(rev, config.nick) for rev in revisions],
                        #get_category=lambda run1, run2: run1.get("domain"),
                    #),
                    #outfile="{}-{}-{}.png".format(exp.name, attribute, config.nick)
                #)

    exp()

main(revisions=['issue682-base', 'issue682-v1'])



__all__ = ["ParseError", "parse_nested_list"]

class ParseError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return self.value

# Basic functions for parsing PDDL (Lisp) files.
def parse_nested_list(input_file):
    tokens = tokenize(input_file)
    next_token = next(tokens)
    if next_token != "(":
        raise ParseError("Expected '(', got %s." % next_token)
    result = list(parse_list_aux(tokens))
    for tok in tokens:  # Check that generator is exhausted.
        raise ParseError("Unexpected token: %s." % tok)
    return result

def tokenize(input):
    for line in input:
        line = line.split(";", 1)[0]  # Strip comments.
        try:
            line.encode("ascii")
        except UnicodeEncodeError:
            raise ParseError("Non-ASCII character outside comment: %s" %
                             line[0:-1])
        line = line.replace("(", " ( ").replace(")", " ) ").replace("?", " ?")
        for token in line.split():
            yield token.lower()

def parse_list_aux(tokenstream):
    # Leading "(" has already been swallowed.
    while True:
        try:
            token = next(tokenstream)
        except StopIteration:
            raise ParseError("Missing ')'")
        if token == ")":
            return
        elif token == "(":
            yield list(parse_list_aux(tokenstream))
        else:
            yield token



from itertools import count
import os

def _try_remove(f):
    try:
        os.remove(f)
    except OSError:
        return False
    return True

def cleanup_temporary_files(args):
    _try_remove(args.sas_file)
    _try_remove(args.plan_file)

    for i in count(1):
        if not _try_remove("%s.%s" % (args.plan_file, i)):
            break



#! /usr/bin/env python

from standard_experiment import REMOTE, get_exp

from downward import suites
#from lab.reports import Attribute, avg

import os.path

# Set the following variables for the experiment
REPO_NAME = 'fd-issue123'
# revisions, e.g. ['3d6c1ccacdce']
REVISIONS = ['issue123-base']
# suites, e.g. ['gripper:prob01.pddl', 'zenotravel:pfile1'] or suites.suite_satisficing_with_ipc11()
LOCAL_SUITE = ['depot:pfile1']
GRID_SUITE = suites.suite_satisficing_with_ipc11()
# configs, e.g. '--search', 'astar(lmcut())' for config
CONFIGS = {

    'lama-2011': [
        "--if-unit-cost",
        "--heuristic",
        "hlm,hff=lm_ff_syn(lm_rhw(reasonable_orders=true))",
        "--search", "iterated(["
        "                 lazy_greedy([hff,hlm],preferred=[hff,hlm]),"
        "                 lazy_wastar([hff,hlm],preferred=[hff,hlm],w=5),"
        "                 lazy_wastar([hff,hlm],preferred=[hff,hlm],w=3),"
        "                 lazy_wastar([hff,hlm],preferred=[hff,hlm],w=2),"
        "                 lazy_wastar([hff,hlm],preferred=[hff,hlm],w=1)"
        "                 ],repeat_last=true,continue_on_fail=true)",
        "--if-non-unit-cost",
        "--heuristic",
        "hlm1,hff1=lm_ff_syn(lm_rhw(reasonable_orders=true,"
        "                           lm_cost_type=one,cost_type=one))",
        "--heuristic",
        "hlm2,hff2=lm_ff_syn(lm_rhw(reasonable_orders=true,"
        "                           lm_cost_type=plusone,cost_type=plusone))",
        "--search", "iterated(["
        "                 lazy_greedy([hff1,hlm1],preferred=[hff1,hlm1],"
        "                             cost_type=one,reopen_closed=false),"
        "                 lazy_greedy([hff2,hlm2],preferred=[hff2,hlm2],"
        "                             reopen_closed=false),"
        "                 lazy_wastar([hff2,hlm2],preferred=[hff2,hlm2],w=5),"
        "                 lazy_wastar([hff2,hlm2],preferred=[hff2,hlm2],w=3),"
        "                 lazy_wastar([hff2,hlm2],preferred=[hff2,hlm2],w=2),"
        "                 lazy_wastar([hff2,hlm2],preferred=[hff2,hlm2],w=1)"
        "                 ],repeat_last=true,continue_on_fail=true)",
    ],

    'lama-2011-first-it': [
        "--heuristic",
        "hlm,hff=lm_ff_syn(lm_rhw(reasonable_orders=true,"
        "                         lm_cost_type=one,cost_type=one))",
        "--search",
        "lazy_greedy([hff,hlm],preferred=[hff,hlm],cost_type=one)"
    ],

    'lama-2011-separated': [
        "--if-unit-cost",
        "--heuristic",
        "hlm=lmcount(lm_rhw(reasonable_orders=true),pref=true)",
        "--heuristic",
        "hff=ff()",
        "--search", "iterated(["
        "                 lazy_greedy([hff,hlm],preferred=[hff,hlm]),"
        "                 lazy_wastar([hff,hlm],preferred=[hff,hlm],w=5),"
        "                 lazy_wastar([hff,hlm],preferred=[hff,hlm],w=3),"
        "                 lazy_wastar([hff,hlm],preferred=[hff,hlm],w=2),"
        "                 lazy_wastar([hff,hlm],preferred=[hff,hlm],w=1)"
        "                 ],repeat_last=true,continue_on_fail=true)",
        "--if-non-unit-cost",
        "--heuristic",
        "hlm1=lmcount(lm_rhw(reasonable_orders=true,"
        "                    lm_cost_type=one,cost_type=one),"
        "             pref=true,cost_type=one)",
        "--heuristic",
        "hff1=ff(cost_type=one)",
        "--heuristic",
        "hlm2=lmcount(lm_rhw(reasonable_orders=true,"
        "                    lm_cost_type=plusone,cost_type=plusone),"
        "             pref=true,cost_type=plusone)",
        "--heuristic",
        "hff2=ff(cost_type=plusone)",
        "--search", "iterated(["
        "                 lazy_greedy([hff1,hlm1],preferred=[hff1,hlm1],"
        "                             cost_type=one,reopen_closed=false),"
        "                 lazy_greedy([hff2,hlm2],preferred=[hff2,hlm2],"
        "                             reopen_closed=false),"
        "                 lazy_wastar([hff2,hlm2],preferred=[hff2,hlm2],w=5),"
        "                 lazy_wastar([hff2,hlm2],preferred=[hff2,hlm2],w=3),"
        "                 lazy_wastar([hff2,hlm2],preferred=[hff2,hlm2],w=2),"
        "                 lazy_wastar([hff2,hlm2],preferred=[hff2,hlm2],w=1)"
        "                 ],repeat_last=true,continue_on_fail=true)",
    ],

    'lama-2011-first-it-separated': [
        "--heuristic",
        "hlm=lmcount(lm_rhw(reasonable_orders=true,"
        "                    lm_cost_type=one,cost_type=one),"
        "             pref=true,cost_type=one)",
        "--heuristic",
        "hff=ff(cost_type=one)",
        "--search",
        "lazy_greedy([hff,hlm],preferred=[hff,hlm],cost_type=one)",
    ],
}

# limits, e.g. { 'search_time': 120 }
LIMITS = None
# for 'make debug', set to True.
COMPILATION_OPTION = None #(default: 'release')
# choose any lower priority if whished
PRIORITY = None #(default: 0)


# Do not change anything below here
SCRIPT_PATH = os.path.abspath(__file__)
if REMOTE:
    SUITE = GRID_SUITE
    REPO = os.path.expanduser('~/repos/' + REPO_NAME)
else:
    SUITE = LOCAL_SUITE
    REPO = os.path.expanduser('~/work/' + REPO_NAME)


# Create the experiment. Add parsers, fetchers or reports...
exp = get_exp(script_path=SCRIPT_PATH, repo=REPO, suite=SUITE,
              configs=CONFIGS, revisions=REVISIONS, limits=LIMITS,
              compilation_option=COMPILATION_OPTION, priority=PRIORITY)
exp.add_score_attributes()
exp.add_extra_attributes(['quality'])

REV = REVISIONS[0]
configs_lama = [('%s-lama-2011' % REV, '%s-lama-2011-separated' % REV)]
exp.add_configs_report(compared_configs=configs_lama, name='lama')
configs_lama_first_it = [('%s-lama-2011-first-it' % REV, '%s-lama-2011-first-it-separated' % REV)]
exp.add_configs_report(compared_configs=configs_lama_first_it, name='lama-first-it')
exp.add_absolute_report()

exp()



#! /usr/bin/env python

from lab.parser import Parser

parser = Parser()

parser.add_pattern('sg_construction_time', 'time for root successor generation creation: (.+)s', type=float)
parser.add_pattern('sg_peak_mem_diff', 'peak memory difference for root successor generator creation: (\d+) KB', type=int)

parser.parse()



# -*- coding: utf-8 -*-

from collections import defaultdict, deque
from itertools import chain
import heapq

import sccs

DEBUG = False

class CausalGraph(object):
    """Weighted causal graph used for defining a variable order.

    The causal graph only contains pre->eff edges (in contrast to the
    variant that also has eff<->eff edges).

    The variable order is defined such that removing all edges v->v'
    with v>v' induces an acyclic subgraph of the causal graph. This
    corresponds to the pruning of the causal graph as described in the
    JAIR 2006 Fast Downward paper for the causal graph heuristic. The
    greedy method is based on weighting the edges of the causal graph.
    In this implementation these weights slightly differ from the
    description in the JAIR paper to reproduce the behaviour of the
    original implementation in the preprocessor component of the
    planner.
    """

    def __init__(self, sas_task):
        self.weighted_graph = defaultdict(lambda: defaultdict(int))
        ## var_no -> (var_no -> number)
        self.predecessor_graph = defaultdict(set)
        self.ordering = []

        self.weight_graph_from_ops(sas_task.operators)
        self.weight_graph_from_axioms(sas_task.axioms)

        self.num_variables = len(sas_task.variables.ranges)
        self.goal_map = dict(sas_task.goal.pairs)

    def get_ordering(self):
        if not self.ordering:
            sccs = self.get_strongly_connected_components()
            self.calculate_topological_pseudo_sort(sccs)
        return self.ordering

    def weight_graph_from_ops(self, operators):
        ### A source variable can be processed several times. This was
        ### probably not intended originally but in experiments (cf.
        ### issue26) it performed better than the (clearer) weighting
        ### described in the Fast Downward paper (which would require
        ### a more complicated implementation).
        for op in operators:
            source_vars = [var for (var, value) in op.prevail]
            for var, pre, _, _ in op.pre_post:
                if pre != -1:
                    source_vars.append(var)

            for target, _, _, cond in op.pre_post:
                for source in chain(source_vars, (var for var, _ in cond)):
                    if source != target:
                        self.weighted_graph[source][target] += 1
                        self.predecessor_graph[target].add(source)

    def weight_graph_from_axioms(self, axioms):
        for ax in axioms:
            target = ax.effect[0]
            for source, _ in ax.condition:
                if source != target:
                    self.weighted_graph[source][target] += 1
                    self.predecessor_graph[target].add(source)

    def get_strongly_connected_components(self):
        unweighted_graph = [[] for _ in range(self.num_variables)]
        assert(len(self.weighted_graph) <= self.num_variables)
        for source, target_weights in self.weighted_graph.items():
            unweighted_graph[source] = sorted(target_weights.keys())
        return sccs.get_sccs_adjacency_list(unweighted_graph)

    def calculate_topological_pseudo_sort(self, sccs):
        for scc in sccs:
            if len(scc) > 1:
                # component needs to be turned into acyclic subgraph

                # Compute subgraph induced by scc
                subgraph = defaultdict(list)
                for var in scc:
                    # for each variable in component only list edges inside
                    # component.
                    subgraph_edges = subgraph[var]
                    for target, cost in sorted(self.weighted_graph[var].items()):
                        if target in scc:
                            if target in self.goal_map:
                                subgraph_edges.append((target, 100000 + cost))
                            subgraph_edges.append((target, cost))

                self.ordering.extend(MaxDAG(subgraph, scc).get_result())
            else:
                self.ordering.append(scc[0])

    def calculate_important_vars(self, goal):
        # Note for future refactoring: it is perhaps more idiomatic
        # and efficient to use a set rather than a defaultdict(bool).
        necessary = defaultdict(bool)
        for var, _ in goal.pairs:
            if not necessary[var]:
                necessary[var] = True
                self.dfs(var, necessary)
        return necessary

    def dfs(self, node, necessary):
        stack = [pred for pred in self.predecessor_graph[node]]
        while stack:
            n = stack.pop()
            if not necessary[n]:
                necessary[n] = True
                stack.extend(pred for pred in self.predecessor_graph[n])


class MaxDAG(object):
    """Defines a variable ordering for a SCC of the (weighted) causal
    graph.

    Conceptually, the greedy algorithm successively picks a node with
    minimal cummulated weight of incoming arcs and removes its
    incident edges from the graph until only a single node remains
    (cf. computation of total order of vertices when pruning the
    causal graph in the Fast Downward JAIR 2006 paper).
    """

    def __init__(self, graph, input_order):
        self.weighted_graph = graph
        # input_order is only used to get the same tie-breaking as
        # with the old preprocessor
        self.input_order = input_order

    def get_result(self):
        incoming_weights = defaultdict(int)
        for weighted_edges in self.weighted_graph.values():
            for target, weight in weighted_edges:
                incoming_weights[target] += weight

        weight_to_nodes = defaultdict(deque)
        for node in self.input_order:
            weight = incoming_weights[node]
            weight_to_nodes[weight].append(node)
        weights = list(weight_to_nodes.keys())
        heapq.heapify(weights)

        done = set()
        result = []
        while weights:
            min_key = weights[0]
            min_elem = None
            entries = weight_to_nodes[min_key]
            while (entries and
                (min_elem is None or min_elem in done or
                min_key > incoming_weights[min_elem])):
                min_elem = entries.popleft()
            if not entries:
                del weight_to_nodes[min_key]
                heapq.heappop(weights) # remove min_key from heap
            if min_elem is None or min_elem in done:
                # since we use lazy deletion from the heap weights,
                # there can be weights with a "done" entry in
                # weight_to_nodes
                continue
            done.add(min_elem)
            result.append(min_elem)
            for target, weight in self.weighted_graph[min_elem]:
                if target not in done:
                    weight = weight % 100000
                    if weight == 0:
                        continue
                    old_in_weight = incoming_weights[target]
                    new_in_weight = old_in_weight - weight
                    incoming_weights[target] = new_in_weight

                    # add new entry to weight_to_nodes
                    if new_in_weight not in weight_to_nodes:
                        heapq.heappush(weights, new_in_weight)
                    weight_to_nodes[new_in_weight].append(target)
        return result


class VariableOrder(object):
    """Apply a given variable order to a SAS task."""
    def __init__(self, ordering):
        """Ordering is a list of variable numbers in the desired order.

        If a variable does not occur in the ordering, it is removed
        from the task.
        """
        self.ordering = ordering
        self.new_var = dict((v, i) for i, v in enumerate(ordering))

    def apply_to_task(self, sas_task):
        self._apply_to_variables(sas_task.variables)
        self._apply_to_init(sas_task.init)
        self._apply_to_goal(sas_task.goal)
        self._apply_to_mutexes(sas_task.mutexes)
        self._apply_to_operators(sas_task.operators)
        self._apply_to_axioms(sas_task.axioms)
        if DEBUG:
            sas_task.validate()

    def _apply_to_variables(self, variables):
        ranges = []
        layers = []
        names = []
        for index, var in enumerate(self.ordering):
            ranges.append(variables.ranges[var])
            layers.append(variables.axiom_layers[var])
            names.append(variables.value_names[var])
        variables.ranges = ranges
        variables.axiom_layers = layers
        variables.value_names = names

    def _apply_to_init(self, init):
        init.values = [init.values[var] for var in self.ordering]

    def _apply_to_goal(self, goal):
        goal.pairs = sorted((self.new_var[var], val)
                            for var, val in goal.pairs
                            if var in self.new_var)

    def _apply_to_mutexes(self, mutexes):
        new_mutexes = []
        for group in mutexes:
            facts = [(self.new_var[var], val) for var, val in group.facts
                     if var in self.new_var]
            if facts and len(set(var for var, _ in facts)) > 1:
                group.facts = facts
                new_mutexes.append(group)
        print("%s of %s mutex groups necessary." % (len(new_mutexes),
                                                    len(mutexes)))
        mutexes[:] = new_mutexes

    def _apply_to_operators(self, operators):
        new_ops = []
        for op in operators:
            pre_post = []
            for eff_var, pre, post, cond in op.pre_post:
                if eff_var in self.new_var:
                    new_cond = list((self.new_var[var], val)
                                    for var, val in cond
                                    if var in self.new_var)
                    pre_post.append(
                        (self.new_var[eff_var], pre, post, new_cond))
            if pre_post:
                op.pre_post = pre_post
                op.prevail = [(self.new_var[var], val)
                              for var, val in op.prevail
                              if var in self.new_var]
                new_ops.append(op)
        print("%s of %s operators necessary." % (len(new_ops),
                                                 len(operators)))
        operators[:] = new_ops

    def _apply_to_axioms(self, axioms):
        new_axioms = []
        for ax in axioms:
            eff_var, eff_val = ax.effect
            if eff_var in self.new_var:
                ax.condition = [(self.new_var[var], val)
                                for var, val in ax.condition
                                if var in self.new_var]
                ax.effect = (self.new_var[eff_var], eff_val)
                new_axioms.append(ax)
        print("%s of %s axiom rules necessary." % (len(new_axioms),
                                                   len(axioms)))
        axioms[:] = new_axioms


def find_and_apply_variable_order(sas_task, reorder_vars=True,
                                  filter_unimportant_vars=True):
    if reorder_vars or filter_unimportant_vars:
        cg = CausalGraph(sas_task)
        if reorder_vars:
            order = cg.get_ordering()
        else:
            order = list(range(len(sas_task.variables.ranges)))
        if filter_unimportant_vars:
            necessary = cg.calculate_important_vars(sas_task.goal)
            print("%s of %s variables necessary." % (len(necessary),
                                                     len(order)))
            order = [var for var in order if necessary[var]]
        VariableOrder(order).apply_to_task(sas_task)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import itertools
import os

from lab.environments import LocalEnvironment, MaiaEnvironment

from downward.reports.compare import ComparativeReport

import common_setup
from common_setup import IssueConfig, IssueExperiment, RelativeScatterPlotReport


DIR = os.path.dirname(os.path.abspath(__file__))
BENCHMARKS_DIR = os.environ["DOWNWARD_BENCHMARKS"]
REVISIONS = ["issue693-v7-base", "issue693-v7"]
BUILDS = ["release32", "release64"]
CONFIGS = [
    IssueConfig(
        "lama-first-{build}".format(**locals()),
        [],
        build_options=[build],
        driver_options=["--build", build, "--alias", "lama-first"])
    for build in BUILDS
]
SUITE = common_setup.DEFAULT_SATISFICING_SUITE
ENVIRONMENT = MaiaEnvironment(
    priority=0, email="jendrik.seipp@unibas.ch")

if common_setup.is_test_run():
    SUITE = IssueExperiment.DEFAULT_TEST_SUITE
    ENVIRONMENT = LocalEnvironment(processes=1)

exp = IssueExperiment(
    revisions=REVISIONS,
    configs=CONFIGS,
    environment=ENVIRONMENT,
)
exp.add_suite(BENCHMARKS_DIR, SUITE)
exp.add_comparison_table_step()
exp.run_steps()



# -*- coding: utf-8 -*-

from collections import defaultdict
import os

from lab import tools

from matplotlib import ticker

from downward.reports.scatter import ScatterPlotReport
from downward.reports.plot import PlotReport, Matplotlib, MatplotlibPlot


# TODO: handle outliers

# TODO: this is mostly copied from ScatterMatplotlib (scatter.py)
class RelativeScatterMatplotlib(Matplotlib):
    @classmethod
    def _plot(cls, report, axes, categories, styles):
        # Display grid
        axes.grid(b=True, linestyle='-', color='0.75')

        has_points = False
        # Generate the scatter plots
        for category, coords in sorted(categories.items()):
            X, Y = zip(*coords)
            axes.scatter(X, Y, s=42, label=category, **styles[category])
            if X and Y:
                has_points = True

        if report.xscale == 'linear' or report.yscale == 'linear':
            plot_size = report.missing_val * 1.01
        else:
            plot_size = report.missing_val * 1.25

        # make 5 ticks above and below 1
        yticks = []
        tick_step = report.ylim_top**(1/5.0)
        for i in xrange(-5, 6):
            yticks.append(tick_step**i)
        axes.set_yticks(yticks)
        axes.get_yaxis().set_major_formatter(ticker.ScalarFormatter())

        axes.set_xlim(report.xlim_left or -1, report.xlim_right or plot_size)
        axes.set_ylim(report.ylim_bottom or -1, report.ylim_top or plot_size)

        for axis in [axes.xaxis, axes.yaxis]:
            MatplotlibPlot.change_axis_formatter(axis,
                                report.missing_val if report.show_missing else None)
        return has_points



class RelativeScatterPlotReport(ScatterPlotReport):
    """
    Generate a scatter plot that shows how a specific attribute in two
    configurations. The attribute value in config 1 is shown on the
    x-axis and the relation to the value in config 2 on the y-axis.
    """

    def __init__(self, show_missing=True, get_category=None, **kwargs):
        ScatterPlotReport.__init__(self, show_missing, get_category, **kwargs)
        if self.output_format == 'tex':
            raise "not supported"
        else:
            self.writer = RelativeScatterMatplotlib

    def _fill_categories(self, runs):
        # We discard the *runs* parameter.
        # Map category names to value tuples
        categories = defaultdict(list)
        self.ylim_bottom = 2
        self.ylim_top = 0.5
        self.xlim_left = float("inf")
        for (domain, problem), runs in self.problem_runs.items():
            if len(runs) != 2:
                continue
            run1, run2 = runs
            assert (run1['config'] == self.configs[0] and
                    run2['config'] == self.configs[1])
            val1 = run1.get(self.attribute)
            val2 = run2.get(self.attribute)
            if val1 is None or val2 is None:
                continue
            category = self.get_category(run1, run2)
            assert val1 > 0, (domain, problem, self.configs[0], val1)
            assert val2 > 0, (domain, problem, self.configs[1], val2)
            x = val1
            y = val2 / float(val1)

            categories[category].append((x, y))

            self.ylim_top = max(self.ylim_top, y)
            self.ylim_bottom = min(self.ylim_bottom, y)
            self.xlim_left = min(self.xlim_left, x)

        # center around 1
        if self.ylim_bottom < 1:
            self.ylim_top = max(self.ylim_top, 1 / float(self.ylim_bottom))
        if self.ylim_top > 1:
            self.ylim_bottom = min(self.ylim_bottom, 1 / float(self.ylim_top))
        return categories

    def _set_scales(self, xscale, yscale):
        # ScatterPlots use log-scaling on the x-axis by default.
        default_xscale = 'log'
        if self.attribute and self.attribute in self.LINEAR:
            default_xscale = 'linear'
        PlotReport._set_scales(self, xscale or default_xscale, 'log')



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import itertools
import os

from lab.environments import LocalEnvironment, MaiaEnvironment

from downward.reports.compare import ComparativeReport

import common_setup
from common_setup import IssueConfig, IssueExperiment, RelativeScatterPlotReport


DIR = os.path.dirname(os.path.abspath(__file__))
BENCHMARKS_DIR = os.environ["DOWNWARD_BENCHMARKS"]
REVISIONS = ["issue213-v1", "issue213-v5", "issue213-v6"]
BUILDS = ["release32", "release64"]
SEARCHES = [
    ("blind", "astar(blind())"),
]
CONFIGS = [
    IssueConfig(
        "{nick}-{build}".format(**locals()),
        ["--search", search],
        build_options=[build],
        driver_options=["--build", build])
    for nick, search in SEARCHES
    for build in BUILDS
]
SUITE = common_setup.DEFAULT_OPTIMAL_SUITE
ENVIRONMENT = MaiaEnvironment(
    priority=0, email="jendrik.seipp@unibas.ch")

if common_setup.is_test_run():
    SUITE = IssueExperiment.DEFAULT_TEST_SUITE
    ENVIRONMENT = LocalEnvironment(processes=1)

exp = IssueExperiment(
    revisions=REVISIONS,
    configs=CONFIGS,
    environment=ENVIRONMENT,
)
exp.add_suite(BENCHMARKS_DIR, SUITE)
exp.add_command('run-custom-parser', [os.path.join(DIR, 'custom-parser.py')])
exp.add_absolute_report_step()

attributes = [
    "coverage", "error", "expansions_until_last_jump", "memory",
    "score_memory", "total_time", "score_total_time",
    "hash_set_load_factor", "hash_set_resizings"]

# Compare revisions.
# lmcut-base-32 vs. lmcut-v1-32 vs. lmcut-v3-32
# lmcut-base-64 vs. lmcut-v1-64 vs. lmcut-v3-64
for build in BUILDS:
    for rev1, rev2 in itertools.combinations(REVISIONS, 2):
        algorithm_pairs = [
            ("{rev1}-{config_nick}-{build}".format(**locals()),
             "{rev2}-{config_nick}-{build}".format(**locals()),
             "Diff ({config_nick}-{build})".format(**locals()))
            for config_nick, search in SEARCHES]
        exp.add_report(
            ComparativeReport(algorithm_pairs, attributes=attributes),
            name="issue213-{rev1}-vs-{rev2}-{build}".format(**locals()))

# Compare builds.
# lmcut-base-32 vs. lmcut-base-64
# lmcut-v1-32 vs. lmcut-v1-64
# lmcut-v3-32 vs. lmcut v3-64
for build1, build2 in itertools.combinations(BUILDS, 2):
    for rev in REVISIONS:
        algorithm_pairs = [
            ("{rev}-{config_nick}-{build1}".format(**locals()),
             "{rev}-{config_nick}-{build2}".format(**locals()),
             "Diff ({config_nick}-{rev})".format(**locals()))
            for config_nick, search in SEARCHES]
        exp.add_report(
            ComparativeReport(algorithm_pairs, attributes=attributes),
            name="issue213-{build1}-vs-{build2}-{rev}".format(**locals()))

for attribute in ["total_time", "memory"]:
    exp.add_report(
        RelativeScatterPlotReport(
            attributes=[attribute],
            filter_algorithm=["issue213-v1-blind-release64", "issue213-v4-blind-release64"]),
        name="issue213-relative-scatter-blind-m64-v1-vs-v4-{}".format(attribute))

exp.run_steps()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from common_setup import IssueConfig, IssueExperiment
import suites


configs = [
    IssueConfig(
        func,
        ["--search", "astar({}([ipdb(max_time=5),diverse_potentials(),all_states_potential(),initial_state_potential(),sample_based_potentials()]))".format(func)])
    for func in ["max", "ipc_max"]
]
revisions = ["8f1563b36fc7"]

exp = IssueExperiment(
    revisions=revisions,
    configs=configs,
    suite=suites.suite_optimal_strips(),
    test_suite=["depot:pfile1"],
    email="jendrik.seipp@unibas.ch",
)

exp.add_absolute_report_step()

exp()



import os.path
import subprocess

DIR = os.path.dirname(os.path.abspath(__file__))
TRANSLATE_DIR = os.path.dirname(DIR)
REPO = os.path.abspath(os.path.join(DIR, "..", "..", ".."))
BENCHMARKS = os.path.join(REPO, "misc", "tests", "benchmarks")
DOMAIN = os.path.join(BENCHMARKS, "gripper", "domain.pddl")
PROBLEM = os.path.join(BENCHMARKS, "gripper", "prob01.pddl")
SCRIPTS = [
    "build_model.py",
    "graph.py",
    "instantiate.py",
    "invariant_finder.py",
    "normalize.py",
    "pddl_to_prolog.py",
    "translate.py",
]

def test_scripts():
    for script in SCRIPTS:
        script = os.path.join(TRANSLATE_DIR, script)
        folder, filename = os.path.split(script)
        assert subprocess.check_call(["python", filename, DOMAIN, PROBLEM], cwd=folder) == 0



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import os

from lab.environments import LocalEnvironment, BaselSlurmEnvironment

import common_setup
from common_setup import IssueConfig, IssueExperiment
from relativescatter import RelativeScatterPlotReport


DIR = os.path.dirname(os.path.abspath(__file__))
BENCHMARKS_DIR = os.environ["DOWNWARD_BENCHMARKS"]
REVISIONS = ["issue700-base", "issue700-v1"]
CONFIGS = [
    IssueConfig("h2", ["--search", "astar(hm(2))"]),
    IssueConfig("ipdb", ["--search", "astar(ipdb())"]),
]
SUITE = common_setup.DEFAULT_OPTIMAL_SUITE
ENVIRONMENT = BaselSlurmEnvironment(
    email="florian.pommerening@unibas.ch",
    export=["PATH", "DOWNWARD_BENCHMARKS"])

if common_setup.is_test_run():
    SUITE = IssueExperiment.DEFAULT_TEST_SUITE
    ENVIRONMENT = LocalEnvironment(processes=1)

exp = IssueExperiment(
    revisions=REVISIONS,
    configs=CONFIGS,
    environment=ENVIRONMENT,
)
exp.add_suite(BENCHMARKS_DIR, SUITE)

exp.add_absolute_report_step()
exp.add_comparison_table_step()

for attr in ["total_time", "search_time", "memory"]:
    for rev1, rev2 in [("base", "v1")]:
        for config_nick in ["h2", "ipdb"]:
            exp.add_report(RelativeScatterPlotReport(
                attributes=[attr],
                filter_algorithm=["issue700-%s-%s" % (rev1, config_nick),
                                  "issue700-%s-%s" % (rev2, config_nick)],
                get_category=lambda r1, r2: r1["domain"],
            ),
            outfile="issue700-%s-%s-%s-%s.png" % (config_nick, attr, rev1, rev2))

exp.run_steps()



import os
import sys
import platform
import subprocess
from setuptools import setup
from setuptools import find_packages
from wheel.bdist_wheel import bdist_wheel

PACKAGE_NAME = 'downward_dlr'


def get_readme():
    return open(os.path.join(os.path.dirname(__file__), 'README'), 'r').read()


class BuildFastDownward(bdist_wheel):

    def run(self):
        cur_dir = os.path.dirname(os.path.realpath(__file__))
        package_dir = os.path.join(cur_dir, PACKAGE_NAME)
        build_command = str(os.path.join(package_dir, 'build.py'))
        print ("Building The Fast-Downward Planning System...")
        build_process = subprocess.Popen([build_command], cwd=package_dir,
                                         stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        line = build_process.stdout.readline()
        encoding = "utf-8" if sys.stdout.encoding is None else sys.stdout.encoding
        while line:
              sys.stdout.write(line.decode(encoding))
              line = build_process.stdout.readline()
 
        line = build_process.stderr.readline()
        while line:
              sys.stderr.write(line.decode(encoding))
              line = build_process.stderr.readline()

        bdist_wheel.run(self)


setup(
    name=PACKAGE_NAME,
    packages=find_packages(),
    include_package_data=True,
    cmdclass={'bdist_wheel': BuildFastDownward},
    entry_points={'console_scripts': ['fast-downward=' + PACKAGE_NAME + '.downward_dlr:dwdlr_main']},
    version='0.0.4',
    author='Christoph Suerig',
    author_email='christoph.suerig@dlr.de',
    license='GNU General Public License Version 3',
    description=PACKAGE_NAME + ' is a fork of the Fast-Downward Planning System (www.fast-downward.org)',
    long_description=get_readme(),
    keywords='semantic planning pddl fast downward planner fast-downward domain-independent',
    classifiers=[
        'Programming Language :: Python :: 2.7',
        'Programming Language :: C++',
        'Environment :: Console',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Topic :: Scientific/Engineering'
    ]







)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from downward import suites
from relativescatter import RelativeScatterPlotReport
import common_setup


REVS = ["issue582-base", "issue582-v1"]
SUITE = suites.suite_optimal_with_ipc11()

CONFIGS = {
    "astar_lmcut": [
        "--search",
        "astar(lmcut())"],
}

exp = common_setup.IssueExperiment(
    search_revisions=REVS,
    configs=CONFIGS,
    suite=SUITE,
    )

exp.add_report(
    RelativeScatterPlotReport(
        attributes=["total_time"],
        get_category=lambda run1, run2: run1.get("domain"),
    ),
    outfile='issue582_base_v1_total_time.png'
)

exp.add_comparison_table_step()

exp()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function

import sys
import itertools

import pddl
import timers
from functools import reduce

def convert_rules(prog):
    RULE_TYPES = {
        "join": JoinRule,
        "product": ProductRule,
        "project": ProjectRule,
        }
    result = []
    for rule in prog.rules:
        RuleType = RULE_TYPES[rule.type]
        new_effect, new_conditions = variables_to_numbers(
            rule.effect, rule.conditions)
        rule = RuleType(new_effect, new_conditions)
        rule.validate()
        result.append(rule)
    return result

def variables_to_numbers(effect, conditions):
    new_effect_args = list(effect.args)
    rename_map = {}
    for i, arg in enumerate(effect.args):
        if arg[0] == "?":
            rename_map[arg] = i
            new_effect_args[i] = i
    new_effect = pddl.Atom(effect.predicate, new_effect_args)

    # There are three possibilities for arguments in conditions:
    # 1. They are variables that occur in the effect. In that case,
    #    they are replaced by the corresponding position in the
    #    effect, as indicated by the rename_map.
    # 2. They are constants. In that case, the unifier must guarantee
    #    that they are matched appropriately. In that case, they are
    #    not modified (remain strings denoting objects).
    # 3. They are variables that don't occur in the effect (are
    #    projected away). This is only allowed in projection rules.
    #    Such arguments are also not modified (remain "?x" strings).

    new_conditions = []
    for cond in conditions:
        new_cond_args = [rename_map.get(arg, arg) for arg in cond.args]
        new_conditions.append(pddl.Atom(cond.predicate, new_cond_args))
    return new_effect, new_conditions

class BuildRule:
    def prepare_effect(self, new_atom, cond_index):
        effect_args = list(self.effect.args)
        cond = self.conditions[cond_index]
        for var_no, obj in zip(cond.args, new_atom.args):
            if isinstance(var_no, int):
                effect_args[var_no] = obj
        return effect_args
    def __str__(self):
        return "%s :- %s" % (self.effect, ", ".join(map(str, self.conditions)))
    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self)

class JoinRule(BuildRule):
    def __init__(self, effect, conditions):
        self.effect = effect
        self.conditions = conditions
        left_args = conditions[0].args
        right_args = conditions[1].args
        left_vars = set([var for var in left_args if isinstance(var, int)])
        right_vars = set([var for var in right_args if isinstance(var, int)])
        common_vars = sorted(left_vars & right_vars)
        self.common_var_positions = [
            [args.index(var) for var in common_vars]
            for args in (list(left_args), list(right_args))]
        self.atoms_by_key = ({}, {})
    def validate(self):
        assert len(self.conditions) == 2, self
        left_args = self.conditions[0].args
        right_args = self.conditions[1].args
        eff_args = self.effect.args
        left_vars = set([v for v in left_args
                         if isinstance(v, int) or v[0] == "?"])
        right_vars = set([v for v in right_args
                          if isinstance(v, int) or v[0] == "?"])
        eff_vars = set([v for v in eff_args
                        if isinstance(v, int) or v[0] == "?"])
        assert left_vars & right_vars, self
        assert (left_vars | right_vars) == (left_vars & right_vars) | eff_vars, self
    def update_index(self, new_atom, cond_index):
        ordered_common_args = [
            new_atom.args[position]
            for position in self.common_var_positions[cond_index]]
        key = tuple(ordered_common_args)
        self.atoms_by_key[cond_index].setdefault(key, []).append(new_atom)
    def fire(self, new_atom, cond_index, enqueue_func):
        effect_args = self.prepare_effect(new_atom, cond_index)
        ordered_common_args = [
            new_atom.args[position]
            for position in self.common_var_positions[cond_index]]
        key = tuple(ordered_common_args)
        other_cond_index = 1 - cond_index
        other_cond = self.conditions[other_cond_index]
        for atom in self.atoms_by_key[other_cond_index].get(key, []):
            for var_no, obj in zip(other_cond.args, atom.args):
                if isinstance(var_no, int):
                    effect_args[var_no] = obj
            enqueue_func(self.effect.predicate, effect_args)

class ProductRule(BuildRule):
    def __init__(self, effect, conditions):
        self.effect = effect
        self.conditions = conditions
        self.atoms_by_index = [[] for c in self.conditions]
        self.empty_atom_list_no = len(self.conditions)
    def validate(self):
        assert len(self.conditions) >= 2, self
        cond_vars = [set([v for v in cond.args
                          if isinstance(v, int) or v[0] == "?"])
                     for cond in self.conditions]
        all_cond_vars = reduce(set.union, cond_vars)
        eff_vars = set([v for v in self.effect.args
                        if isinstance(v, int) or v[0] == "?"])
        assert len(all_cond_vars) == len(eff_vars), self
        assert len(all_cond_vars) == sum([len(c) for c in cond_vars])
    def update_index(self, new_atom, cond_index):
        atom_list = self.atoms_by_index[cond_index]
        if not atom_list:
            self.empty_atom_list_no -= 1
        atom_list.append(new_atom)

    def _get_bindings(self, atom, cond):
        return [(var_no, obj) for var_no, obj in zip(cond.args, atom.args)
                if isinstance(var_no, int)]

    def fire(self, new_atom, cond_index, enqueue_func):
        if self.empty_atom_list_no:
            return

        # Binding: a (var_no, object) pair
        # Bindings: List-of(Binding)
        # BindingsFactor: List-of(Bindings)
        # BindingsFactors: List-of(BindingsFactor)
        bindings_factors = []
        for pos, cond in enumerate(self.conditions):
            if pos == cond_index:
                continue
            atoms = self.atoms_by_index[pos]
            assert atoms, "if we have no atoms, this should never be called"
            factor = [self._get_bindings(atom, cond) for atom in atoms]
            bindings_factors.append(factor)

        eff_args = self.prepare_effect(new_atom, cond_index)

        for bindings_list in itertools.product(*bindings_factors):
            bindings = itertools.chain(*bindings_list)
            for var_no, obj in bindings:
                eff_args[var_no] = obj
            enqueue_func(self.effect.predicate, eff_args)


class ProjectRule(BuildRule):
    def __init__(self, effect, conditions):
        self.effect = effect
        self.conditions = conditions
    def validate(self):
        assert len(self.conditions) == 1
    def update_index(self, new_atom, cond_index):
        pass
    def fire(self, new_atom, cond_index, enqueue_func):
        effect_args = self.prepare_effect(new_atom, cond_index)
        enqueue_func(self.effect.predicate, effect_args)

class Unifier:
    def __init__(self, rules):
        self.predicate_to_rule_generator = {}
        for rule in rules:
            for i, cond in enumerate(rule.conditions):
                self._insert_condition(rule, i)
    def unify(self, atom):
        result = []
        generator = self.predicate_to_rule_generator.get(atom.predicate)
        if generator:
            generator.generate(atom, result)
        return result
    def _insert_condition(self, rule, cond_index):
        condition = rule.conditions[cond_index]
        root = self.predicate_to_rule_generator.get(condition.predicate)
        if not root:
            root = LeafGenerator()
        constant_arguments = [
            (arg_index, arg)
            for (arg_index, arg) in enumerate(condition.args)
            if not isinstance(arg, int) and arg[0] != "?"]
        newroot = root._insert(constant_arguments, (rule, cond_index))
        self.predicate_to_rule_generator[condition.predicate] = newroot
    def dump(self):
        predicates = sorted(self.predicate_to_rule_generator)
        print("Unifier:")
        for pred in predicates:
            print("    %s:" % pred)
            rule_gen = self.predicate_to_rule_generator[pred]
            rule_gen.dump("    " * 2)

class LeafGenerator:
    index = sys.maxsize
    def __init__(self):
        self.matches = []
    def empty(self):
        return not self.matches
    def generate(self, atom, result):
        result += self.matches
    def _insert(self, args, value):
        if not args:
            self.matches.append(value)
            return self
        else:
            root = LeafGenerator()
            root.matches.append(value)
            for arg_index, arg in args[::-1]:
                new_root = MatchGenerator(arg_index, LeafGenerator())
                new_root.match_generator[arg] = root
                root = new_root
            root.matches = self.matches # can be swapped in C++
            return root
    def dump(self, indent):
        for match in self.matches:
            print("%s%s" % (indent, match))

class MatchGenerator:
    def __init__(self, index, next):
        self.index = index
        self.matches = []
        self.match_generator = {}
        self.next = next
    def empty(self):
        return False
    def generate(self, atom, result):
        result += self.matches
        generator = self.match_generator.get(atom.args[self.index])
        if generator:
            generator.generate(atom, result)
        self.next.generate(atom, result)
    def _insert(self, args, value):
        if not args:
            self.matches.append(value)
            return self
        else:
            arg_index, arg = args[0]
            if self.index < arg_index:
                self.next = self.next._insert(args, value)
                return self
            elif self.index > arg_index:
                new_parent = MatchGenerator(arg_index, self)
                new_branch = LeafGenerator()._insert(args[1:], value)
                new_parent.match_generator[arg] = new_branch
                return new_parent
            else:
                branch_generator = self.match_generator.get(arg)
                if not branch_generator:
                    branch_generator = LeafGenerator()
                self.match_generator[arg] = branch_generator._insert(
                    args[1:], value)
                return self
    def dump(self, indent):
        for match in self.matches:
            print("%s%s" % (indent, match))
        for key in sorted(self.match_generator.keys()):
            print("%sargs[%s] == %s:" % (indent, self.index, key))
            self.match_generator[key].dump(indent + "    ")
        if not self.next.empty():
            assert isinstance(self.next, MatchGenerator)
            print("%s[*]" % indent)
            self.next.dump(indent + "    ")

class Queue:
    def __init__(self, atoms):
        self.queue = atoms
        self.queue_pos = 0
        self.enqueued = set([(atom.predicate,) + tuple(atom.args)
                             for atom in self.queue])
        self.num_pushes = len(atoms)
    def __bool__(self):
        return self.queue_pos < len(self.queue)
    __nonzero__ = __bool__
    def push(self, predicate, args):
        self.num_pushes += 1
        eff_tuple = (predicate,) + tuple(args)
        if eff_tuple not in self.enqueued:
            self.enqueued.add(eff_tuple)
            self.queue.append(pddl.Atom(predicate, list(args)))
    def pop(self):
        result = self.queue[self.queue_pos]
        self.queue_pos += 1
        return result

def compute_model(prog):
    with timers.timing("Preparing model"):
        rules = convert_rules(prog)
        unifier = Unifier(rules)
        # unifier.dump()
        fact_atoms = sorted(fact.atom for fact in prog.facts)
        queue = Queue(fact_atoms)

    print("Generated %d rules." % len(rules))
    with timers.timing("Computing model"):
        relevant_atoms = 0
        auxiliary_atoms = 0
        while queue:
            next_atom = queue.pop()
            pred = next_atom.predicate
            if isinstance(pred, str) and "$" in pred:
                auxiliary_atoms += 1
            else:
                relevant_atoms += 1
            matches = unifier.unify(next_atom)
            for rule, cond_index in matches:
                rule.update_index(next_atom, cond_index)
                rule.fire(next_atom, cond_index, queue.push)
    print("%d relevant atoms" % relevant_atoms)
    print("%d auxiliary atoms" % auxiliary_atoms)
    print("%d final queue length" % len(queue.queue))
    print("%d total queue pushes" % queue.num_pushes)
    return queue.queue

if __name__ == "__main__":
    import pddl_parser
    import normalize
    import pddl_to_prolog

    print("Parsing...")
    task = pddl_parser.open()
    print("Normalizing...")
    normalize.normalize(task)
    print("Writing rules...")
    prog = pddl_to_prolog.translate(task)

    model = compute_model(prog)
    for atom in model:
        print(atom)
    print("%d atoms" % len(model))



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from downward import suites

from common_setup_no_benchmarks import IssueConfig, IssueExperiment, get_script_dir
from relativescatter import RelativeScatterPlotReport

import os

def main(revisions=None):
    exp = IssueExperiment(benchmarks_dir=".", suite=[])

    exp.add_fetcher(
        os.path.join(get_script_dir(), "data", "issue627-v3-eval"),
        filter=lambda(run): "base" not in run["config"],
    )
    exp.add_fetcher(
        os.path.join(get_script_dir(), "data", "issue627-v4-eval"),
        filter=lambda(run): "base" not in run["config"],
    )

    for config_nick in ['astar-blind', 'astar-lmcut', 'astar-ipdb', 'astar-cegar-original', 'astar-cegar-lm-goals']:
        exp.add_report(
            RelativeScatterPlotReport(
                attributes=["memory"],
                filter_config=["issue627-v3-%s" % config_nick,
                               "issue627-v4-%s" % config_nick],
                get_category=lambda run1, run2: run1.get("domain"),
            ),
            outfile='issue627_v3_v4_memory_%s.png' % config_nick
        )

        exp.add_report(
            RelativeScatterPlotReport(
                attributes=["total_time"],
                filter_config=["issue627-v3-%s" % config_nick,
                               "issue627-v4-%s" % config_nick],
                get_category=lambda run1, run2: run1.get("domain"),
            ),
            outfile='issue627_v3_v4_total_time_%s.png' % config_nick
        )

    exp()

main(revisions=['issue627-v3', 'issue627-v4'])



# -*- coding: utf-8 -*-

from __future__ import print_function

import os
import re

from . import returncodes


DRIVER_DIR = os.path.abspath(os.path.dirname(__file__))
REPO_ROOT_DIR = os.path.dirname(DRIVER_DIR)
BUILDS_DIR = os.path.join(REPO_ROOT_DIR, "builds")


def get_elapsed_time():
    """
    Return the CPU time taken by the python process and its child
    processes.
    """
    if os.name == "nt":
        # The child time components of os.times() are 0 on Windows.
        raise NotImplementedError("cannot use get_elapsed_time() on Windows")
    return sum(os.times()[:4])


def find_domain_filename(task_filename):
    """
    Find domain filename for the given task using automatic naming rules.
    """
    dirname, basename = os.path.split(task_filename)

    domain_basenames = [
        "domain.pddl",
        basename[:3] + "-domain.pddl",
        "domain_" + basename,
        "domain-" + basename,
    ]

    for domain_basename in domain_basenames:
        domain_filename = os.path.join(dirname, domain_basename)
        if os.path.exists(domain_filename):
            return domain_filename

    returncodes.exit_with_driver_input_error(
        "Error: Could not find domain file using automatic naming rules.")

# Shell-escaping code taken from Python's shlex.quote (missing in Python < 3.3).
_find_unsafe = re.compile(r'[^\w@%+=:,./-]').search

def shell_escape(s):
    """Return a shell-escaped version of the string *s*."""
    if not s:
        return "''"
    if _find_unsafe(s) is None:
        return s

    # Use single quotes, and put single quotes into double quotes.
    # The string $'b is then quoted as '$'"'"'b'.
    return "'" + s.replace("'", "'\"'\"'") + "'"



#! /usr/bin/env python
# -*- coding: utf-8 -*-

'''
Script to test possible eager version of LAMA

'''

import os

from lab.environments import LocalEnvironment, BaselSlurmEnvironment

import common_setup
from common_setup import IssueConfig, IssueExperiment
from relativescatter import RelativeScatterPlotReport

DIR = os.path.dirname(os.path.abspath(__file__))
BENCHMARKS_DIR = os.environ["DOWNWARD_BENCHMARKS"]
REVISIONS = ["issue311"]
CONFIGS = [
    IssueConfig('lama', [], driver_options=["--alias", "seq-sat-lama-2011"]),
    IssueConfig('eager_lama', [
        "--if-unit-cost",
        "--evaluator",
        "hlm=lama_synergy(lm_rhw(reasonable_orders=true))",
        "--evaluator", "hff=ff_synergy(hlm)",
        "--search", """iterated([
        lazy_greedy([hff,hlm],preferred=[hff,hlm]),
        eager_wastar([hff,hlm],preferred=[hff,hlm],w=5),
        eager_wastar([hff,hlm],preferred=[hff,hlm],w=3),
        eager_wastar([hff,hlm],preferred=[hff,hlm],w=2),
        eager_wastar([hff,hlm],preferred=[hff,hlm],w=1)
        ],repeat_last=true,continue_on_fail=true)""",
        "--if-non-unit-cost",
        "--evaluator",
        "hlm1=lama_synergy(lm_rhw(reasonable_orders=true,"
        "                           lm_cost_type=one),transform=adapt_costs(one))",
        "--evaluator", "hff1=ff_synergy(hlm1)",
        "--evaluator",
        "hlm2=lama_synergy(lm_rhw(reasonable_orders=true,"
        "                           lm_cost_type=plusone),transform=adapt_costs(plusone))",
        "--evaluator", "hff2=ff_synergy(hlm2)",
        "--search", """iterated([
        lazy_greedy([hff1,hlm1],preferred=[hff1,hlm1],
        cost_type=one,reopen_closed=false),
        lazy_greedy([hff2,hlm2],preferred=[hff2,hlm2],
        reopen_closed=false),
        eager_wastar([hff2,hlm2],preferred=[hff2,hlm2],w=5),
        eager_wastar([hff2,hlm2],preferred=[hff2,hlm2],w=3),
        eager_wastar([hff2,hlm2],preferred=[hff2,hlm2],w=2),
        eager_wastar([hff2,hlm2],preferred=[hff2,hlm2],w=1)
        ],repeat_last=true,continue_on_fail=true)""",
        "--always"
    ])
]

SUITE = common_setup.DEFAULT_SATISFICING_SUITE
ENVIRONMENT = BaselSlurmEnvironment(
    partition="infai_1",
    export=["PATH", "DOWNWARD_BENCHMARKS"])

if common_setup.is_test_run():
    SUITE = IssueExperiment.DEFAULT_TEST_SUITE
    ENVIRONMENT = LocalEnvironment(processes=1)

exp = IssueExperiment(
    revisions=REVISIONS,
    configs=CONFIGS,
    environment=ENVIRONMENT,
)
exp.add_suite(BENCHMARKS_DIR, SUITE)

exp.add_parser(exp.EXITCODE_PARSER)
exp.add_parser(exp.TRANSLATOR_PARSER)
exp.add_parser(exp.ANYTIME_SEARCH_PARSER)
exp.add_parser(exp.PLANNER_PARSER)

exp.add_step('build', exp.build)
exp.add_step('start', exp.start_runs)
exp.add_fetcher(name='fetch')

exp.add_absolute_report_step()
exp.add_comparison_table_step()

exp.run_steps()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import itertools
import os

from lab.environments import LocalEnvironment, BaselSlurmEnvironment

from downward.reports.compare import ComparativeReport

import common_setup
from common_setup import IssueConfig, IssueExperiment
from relativescatter import RelativeScatterPlotReport

DIR = os.path.dirname(os.path.abspath(__file__))
SCRIPT_NAME = os.path.splitext(os.path.basename(__file__))[0]
BENCHMARKS_DIR = os.environ["DOWNWARD_BENCHMARKS"]
REVISIONS = ["issue901-base", "issue901-v1"]
CONFIGS = [
    IssueConfig("cegar-original", ["--search", "astar(cegar(subtasks=[original()], max_transitions=1M, max_time=infinity))"]),
    IssueConfig("cegar-lm-goals", ["--search", "astar(cegar(subtasks=[landmarks(), goals()], max_transitions=1M, max_time=infinity))"]),
]

SUITE = common_setup.DEFAULT_OPTIMAL_SUITE
ENVIRONMENT = BaselSlurmEnvironment(
    partition="infai_2",
    email="jendrik.seipp@unibas.ch",
    export=["PATH", "DOWNWARD_BENCHMARKS"])

if common_setup.is_test_run():
    SUITE = IssueExperiment.DEFAULT_TEST_SUITE
    ENVIRONMENT = LocalEnvironment(processes=1)

exp = IssueExperiment(
    revisions=REVISIONS,
    configs=CONFIGS,
    environment=ENVIRONMENT,
)
exp.add_suite(BENCHMARKS_DIR, SUITE)

exp.add_parser(exp.EXITCODE_PARSER)
exp.add_parser(exp.SINGLE_SEARCH_PARSER)
exp.add_parser(exp.PLANNER_PARSER)
exp.add_parser(os.path.join(DIR, "parser.py"))

exp.add_step('build', exp.build)
exp.add_step('start', exp.start_runs)
exp.add_fetcher(name='fetch')

REFINEMENT_ATTRIBUTES = [
    "time_for_finding_traces",
    "time_for_finding_flaws",
    "time_for_splitting_states",
]
attributes = (
    IssueExperiment.DEFAULT_TABLE_ATTRIBUTES +
    ["search_start_memory", "init_time", "time_analysis"] +
    ["total_" + attr for attr in REFINEMENT_ATTRIBUTES])
#exp.add_absolute_report_step(attributes=attributes)
exp.add_comparison_table_step(attributes=attributes)
exp.add_scatter_plot_step(relative=True, attributes=["search_time", "total_time"])

exp.run_steps()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import os
import suites
from lab.reports import Attribute, gm

from common_setup import IssueConfig, IssueExperiment
try:
    from relativescatter import RelativeScatterPlotReport
    matplotlib = True
except ImportError:
    print 'matplotlib not available, scatter plots not available'
    matplotlib = False

def main(revisions=None):
    benchmarks_dir=os.path.expanduser('~/projects/downward/benchmarks')
    suite=suites.suite_optimal_strips()

    configs = {
        IssueConfig('astar-blind', ['--search', 'astar(blind())']),
        IssueConfig('astar-blind-sss', ['--search', 'astar(blind(), pruning=stubborn_sets_simple())']),
        IssueConfig('astar-blind-ssec', ['--search', 'astar(blind(), pruning=stubborn_sets_ec())']),
    }

    exp = IssueExperiment(
        benchmarks_dir=benchmarks_dir,
        suite=suite,
        revisions=revisions,
        configs=configs,
        test_suite=['depot:p01.pddl'],
        processes=4,
        email='florian.pommerening@unibas.ch',
    )

    exp.add_comparison_table_step()

    if matplotlib:
        for attribute in ["memory", "total_time"]:
            for config in configs:
                exp.add_report(
                    RelativeScatterPlotReport(
                        attributes=[attribute],
                        filter_config=["{}-{}".format(rev, config.nick) for rev in revisions],
                        get_category=lambda run1, run2: run1.get("domain"),
                    ),
                    outfile="{}-{}-{}.png".format(exp.name, attribute, config.nick)
                )

    exp()

main(revisions=['issue629-base', 'issue629-v1'])



# -*- coding: utf-8 -*-

from __future__ import print_function

import errno
import logging
import os.path
import subprocess
import sys

from . import call
from . import limits
from . import portfolio_runner
from . import returncodes
from . import util
from .plan_manager import PlanManager

#TODO: We might want to turn translate into a module and call it with "python -m translate".
REL_TRANSLATE_PATH = os.path.join("translate", "translate.py")
if os.name == "posix":
    REL_SEARCH_PATH = "downward"
    VALIDATE = "validate"
elif os.name == "nt":
    REL_SEARCH_PATH = "downward.exe"
    VALIDATE = "validate.exe"
else:
    returncodes.exit_with_driver_unsupported_error("Unsupported OS: " + os.name)

def get_executable(build, rel_path):
    # First, consider 'build' to be a path directly to the binaries.
    # The path can be absolute or relative to the current working
    # directory.
    build_dir = build
    if not os.path.exists(build_dir):
        # If build is not a full path to the binaries, it might be the
        # name of a build in our standard directory structure.
        # in this case, the binaries are in
        #   '<repo-root>/builds/<buildname>/bin'.
        build_dir = os.path.join(util.BUILDS_DIR, build, "bin")
        if not os.path.exists(build_dir):
            returncodes.exit_with_driver_input_error(
                "Could not find build '{build}' at {build_dir}. "
                "Please run './build.py {build}'.".format(**locals()))

    abs_path = os.path.join(build_dir, rel_path)
    if not os.path.exists(abs_path):
        returncodes.exit_with_driver_input_error(
            "Could not find '{rel_path}' in build '{build}'. "
            "Please run './build.py {build}'.".format(**locals()))

    return abs_path


def run_translate(args):
    logging.info("Running translator.")
    time_limit = limits.get_time_limit(
        args.translate_time_limit, args.overall_time_limit)
    memory_limit = limits.get_memory_limit(
        args.translate_memory_limit, args.overall_memory_limit)
    translate = get_executable(args.build, REL_TRANSLATE_PATH)
    assert sys.executable, "Path to interpreter could not be found"
    cmd = [sys.executable] + [translate] + args.translate_inputs + args.translate_options

    stderr, returncode = call.get_error_output_and_returncode(
        "translator",
        cmd,
        time_limit=time_limit,
        memory_limit=memory_limit)

    # We collect stderr of the translator and print it here, unless
    # the translator ran out of memory and all output in stderr is
    # related to MemoryError.
    do_print_on_stderr = True
    if returncode == returncodes.TRANSLATE_OUT_OF_MEMORY:
        output_related_to_memory_error = True
        if not stderr:
            output_related_to_memory_error = False
        for line in stderr.splitlines():
            if "MemoryError" not in line:
                output_related_to_memory_error = False
                break
        if output_related_to_memory_error:
            do_print_on_stderr = False

    if do_print_on_stderr and stderr:
        returncodes.print_stderr(stderr)

    if returncode == 0:
        return (0, True)
    elif returncode == 1:
        # Unlikely case that the translator crashed without raising an
        # exception.
        return (returncodes.TRANSLATE_CRITICAL_ERROR, False)
    else:
        # Pass on any other exit code, including in particular signals or
        # exit codes such as running out of memory or time.
        return (returncode, False)


def run_search(args):
    logging.info("Running search (%s)." % args.build)
    time_limit = limits.get_time_limit(
        args.search_time_limit, args.overall_time_limit)
    memory_limit = limits.get_memory_limit(
        args.search_memory_limit, args.overall_memory_limit)
    executable = get_executable(args.build, REL_SEARCH_PATH)

    plan_manager = PlanManager(
        args.plan_file,
        portfolio_bound=args.portfolio_bound,
        single_plan=args.portfolio_single_plan)
    plan_manager.delete_existing_plans()

    if args.portfolio:
        assert not args.search_options
        logging.info("search portfolio: %s" % args.portfolio)
        return portfolio_runner.run(
            args.portfolio, executable, args.search_input, plan_manager,
            time_limit, memory_limit)
    else:
        if not args.search_options:
            returncodes.exit_with_driver_input_error(
                "search needs --alias, --portfolio, or search options")
        if "--help" not in args.search_options:
            args.search_options.extend(["--internal-plan-file", args.plan_file])
        try:
            call.check_call(
                "search",
                [executable] + args.search_options,
                stdin=args.search_input,
                time_limit=time_limit,
                memory_limit=memory_limit)
        except subprocess.CalledProcessError as err:
            # TODO: if we ever add support for SEARCH_PLAN_FOUND_AND_* directly
            # in the planner, this assertion no longer holds. Furthermore, we
            # would need to return (err.returncode, True) if the returncode is
            # in [0..10].
            # Negative exit codes are allowed for passing out signals.
            assert err.returncode >= 10 or err.returncode < 0, "got returncode < 10: {}".format(err.returncode)
            return (err.returncode, False)
        else:
            return (0, True)


def run_validate(args):
    logging.info("Running validate.")

    num_files = len(args.filenames)
    if num_files == 1:
        task, = args.filenames
        domain = util.find_domain_filename(task)
    elif num_files == 2:
        domain, task = args.filenames
    else:
        returncodes.exit_with_driver_input_error("validate needs one or two PDDL input files.")

    plan_files = list(PlanManager(args.plan_file).get_existing_plans())
    if not plan_files:
        print("Not running validate since no plans found.")
        return (0, True)
    validate_inputs = [domain, task] + plan_files

    try:
        call.check_call(
            "validate",
            [VALIDATE] + validate_inputs,
            time_limit=args.validate_time_limit,
            memory_limit=args.validate_memory_limit)
    except OSError as err:
        if err.errno == errno.ENOENT:
            returncodes.exit_with_driver_input_error("Error: {} not found. Is it on the PATH?".format(VALIDATE))
        else:
            returncodes.exit_with_driver_critical_error(err)
    else:
        return (0, True)



from __future__ import print_function

import invariant_finder
import options
import pddl
import timers


DEBUG = False


def expand_group(group, task, reachable_facts):
    result = []
    for fact in group:
        try:
            pos = list(fact.args).index("?X")
        except ValueError:
            result.append(fact)
        else:
            # NOTE: This could be optimized by only trying objects of the correct
            #       type, or by using a unifier which directly generates the
            #       applicable objects. It is not worth optimizing this at this stage,
            #       though.
            for obj in task.objects:
                newargs = list(fact.args)
                newargs[pos] = obj.name
                atom = pddl.Atom(fact.predicate, newargs)
                if atom in reachable_facts:
                    result.append(atom)
    return result

def instantiate_groups(groups, task, reachable_facts):
    return [expand_group(group, task, reachable_facts) for group in groups]

class GroupCoverQueue:
    def __init__(self, groups):
        if groups:
            self.max_size = max([len(group) for group in groups])
            self.groups_by_size = [[] for i in range(self.max_size + 1)]
            self.groups_by_fact = {}
            for group in groups:
                group = set(group) # Copy group, as it will be modified.
                self.groups_by_size[len(group)].append(group)
                for fact in group:
                    self.groups_by_fact.setdefault(fact, []).append(group)
            self._update_top()
        else:
            self.max_size = 0
    def __bool__(self):
        return self.max_size > 1
    __nonzero__ = __bool__
    def pop(self):
        result = list(self.top) # Copy; this group will shrink further.
        if options.use_partial_encoding:
            for fact in result:
                for group in self.groups_by_fact[fact]:
                    group.remove(fact)
        self._update_top()
        return result
    def _update_top(self):
        while self.max_size > 1:
            max_list = self.groups_by_size[self.max_size]
            while max_list:
                candidate = max_list.pop()
                if len(candidate) == self.max_size:
                    self.top = candidate
                    return
                self.groups_by_size[len(candidate)].append(candidate)
            self.max_size -= 1

def choose_groups(groups, reachable_facts):
    queue = GroupCoverQueue(groups)
    uncovered_facts = reachable_facts.copy()
    result = []
    while queue:
        group = queue.pop()
        uncovered_facts.difference_update(group)
        result.append(group)
    print(len(uncovered_facts), "uncovered facts")
    result += [[fact] for fact in uncovered_facts]
    return result

def build_translation_key(groups):
    group_keys = []
    for group in groups:
        group_key = [str(fact) for fact in group]
        if len(group) == 1:
            group_key.append(str(group[0].negate()))
        else:
            group_key.append("<none of those>")
        group_keys.append(group_key)
    return group_keys

def collect_all_mutex_groups(groups, atoms):
    # NOTE: This should be functionally identical to choose_groups
    # when partial_encoding is set to False. Maybe a future
    # refactoring could take that into account.
    all_groups = []
    uncovered_facts = atoms.copy()
    for group in groups:
        uncovered_facts.difference_update(group)
        all_groups.append(group)
    all_groups += [[fact] for fact in uncovered_facts]
    return all_groups

def sort_groups(groups):
    return sorted(sorted(group) for group in groups)

def compute_groups(task, atoms, reachable_action_params):
    groups = invariant_finder.get_groups(task, reachable_action_params)

    with timers.timing("Instantiating groups"):
        groups = instantiate_groups(groups, task, atoms)

    # Sort here already to get deterministic mutex groups.
    groups = sort_groups(groups)
    # TODO: I think that collect_all_mutex_groups should do the same thing
    #       as choose_groups with partial_encoding=False, so these two should
    #       be unified.
    with timers.timing("Collecting mutex groups"):
        mutex_groups = collect_all_mutex_groups(groups, atoms)
    with timers.timing("Choosing groups", block=True):
        groups = choose_groups(groups, atoms)
    groups = sort_groups(groups)
    with timers.timing("Building translation key"):
        translation_key = build_translation_key(groups)

    if DEBUG:
        for group in groups:
            if len(group) >= 2:
                print("{%s}" % ", ".join(map(str, group)))

    return groups, mutex_groups, translation_key



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import os
import os.path

from lab.environments import LocalEnvironment, BaselSlurmEnvironment
from lab.reports import Attribute, geometric_mean

import common_setup
from common_setup import IssueConfig, IssueExperiment
from relativescatter import RelativeScatterPlotReport

DIR = os.path.dirname(os.path.abspath(__file__))
BENCHMARKS_DIR = os.environ["DOWNWARD_BENCHMARKS"]
REVISION_CACHE = os.path.expanduser('~/lab/revision-cache')
REVISIONS = ["issue524-base-v2", "issue524-v2"]
CONFIGS = [
    IssueConfig('lm_hm', [
        '--landmarks', 'l=lm_hm()',
        '--heuristic', 'h=lmcount(l)',
        '--search', 'eager_greedy([h])']),
] + [
    IssueConfig('lm_rhw', [
        '--landmarks', 'l=lm_rhw()',
        '--heuristic', 'h=lmcount(l)',
        '--search', 'eager_greedy([h])']),
] + [
    IssueConfig('lm_zg', [
        '--landmarks', 'l=lm_zg()',
        '--heuristic', 'h=lmcount(l)',
        '--search', 'eager_greedy([h])']),
] + [
    IssueConfig('lm_exhaust', [
        '--landmarks', 'l=lm_exhaust()',
        '--heuristic', 'h=lmcount(l)',
        '--search', 'eager_greedy([h])']),
] + [
    IssueConfig('lm_merged', [
        '--landmarks', 'l1=lm_exhaust()',
        '--landmarks', 'l2=lm_rhw()',
        '--landmarks', 'l=lm_merged([l1, l2])',
        '--heuristic', 'h=lmcount(l)',
        '--search', 'eager_greedy([h])']),
] + [
    IssueConfig(
        "lama-first", [], driver_options=["--alias", "lama-first"])
]
SUITE = common_setup.DEFAULT_OPTIMAL_SUITE
ENVIRONMENT = BaselSlurmEnvironment(
    email="cedric.geissmann@unibas.ch"
)

SUITE = common_setup.DEFAULT_OPTIMAL_SUITE

if common_setup.is_test_run():
    SUITE = IssueExperiment.DEFAULT_TEST_SUITE
    ENVIRONMENT = LocalEnvironment(processes=4)

exp = IssueExperiment(
    revisions=REVISIONS,
    configs=CONFIGS,
    environment=ENVIRONMENT,
    revision_cache=REVISION_CACHE,
)
exp.add_suite(BENCHMARKS_DIR, SUITE)

exp.add_absolute_report_step()
exp.add_comparison_table_step()
exp.add_scatter_plot_step()

exp.run_steps()



# -*- coding: utf-8 -*-

from __future__ import print_function

import itertools

class NegativeClause(object):
    # disjunction of inequalities
    def __init__(self, parts):
        self.parts = parts
        assert len(parts)

    def __str__(self):
        disj = " or ".join(["(%s != %s)" % (v1, v2)
                            for (v1, v2) in self.parts])
        return "(%s)" % disj

    def is_satisfiable(self):
        for part in self.parts:
            if part[0] != part[1]:
                return True
        return False

    def apply_mapping(self, m):
        new_parts = [(m.get(v1, v1), m.get(v2, v2)) for (v1, v2) in self.parts]
        return NegativeClause(new_parts)


class Assignment(object):
    def __init__(self, equalities):
        self.equalities = tuple(equalities)
        # represents a conjunction of expressions ?x = ?y or ?x = d
        # with ?x, ?y being variables and d being a domain value

        self.consistent = None
        self.mapping = None
        self.eq_classes = None

    def __str__(self):
        conj = " and ".join(["(%s = %s)" % (v1, v2)
                            for (v1, v2) in self.equalities])
        return "(%s)" % conj

    def _compute_equivalence_classes(self):
        eq_classes = {}
        for (v1, v2) in self.equalities:
            c1 = eq_classes.setdefault(v1, set([v1]))
            c2 = eq_classes.setdefault(v2, set([v2]))
            if c1 is not c2:
                if len(c2) > len(c1):
                    v1, c1, v2, c2 = v2, c2, v1, c1
                c1.update(c2)
                for elem in c2:
                    eq_classes[elem] = c1
        self.eq_classes = eq_classes

    def _compute_mapping(self):
        if not self.eq_classes:
            self._compute_equivalence_classes()

        # create mapping: each key is mapped to the smallest
        # element in its equivalence class (with objects being
        # smaller than variables)
        mapping = {}
        for eq_class in self.eq_classes.values():
            variables = [item for item in eq_class if item.startswith("?")]
            constants = [item for item in eq_class if not item.startswith("?")]
            if len(constants) >= 2:
                self.consistent = False
                self.mapping = None
                return
            if constants:
                set_val = constants[0]
            else:
                set_val = min(variables)
            for entry in eq_class:
                mapping[entry] = set_val
        self.consistent = True
        self.mapping = mapping

    def is_consistent(self):
        if self.consistent is None:
            self._compute_mapping()
        return self.consistent

    def get_mapping(self):
        if self.consistent is None:
            self._compute_mapping()
        return self.mapping


class ConstraintSystem(object):
    def __init__(self):
        self.combinatorial_assignments = []
        self.neg_clauses = []

    def __str__(self):
        combinatorial_assignments = []
        for comb_assignment in self.combinatorial_assignments:
            disj = " or ".join([str(assig) for assig in comb_assignment])
            disj = "(%s)" % disj
            combinatorial_assignments.append(disj)
        assigs = " and\n".join(combinatorial_assignments)

        neg_clauses = [str(clause) for clause in self.neg_clauses]
        neg_clauses = " and ".join(neg_clauses)
        return assigs + "(" + neg_clauses + ")"

    def _all_clauses_satisfiable(self, assignment):
        mapping = assignment.get_mapping()
        for neg_clause in self.neg_clauses:
            clause = neg_clause.apply_mapping(mapping)
            if not clause.is_satisfiable():
                return False
        return True

    def _combine_assignments(self, assignments):
        new_equalities = []
        for a in assignments:
            new_equalities.extend(a.equalities)
        return Assignment(new_equalities)

    def add_assignment(self, assignment):
        self.add_assignment_disjunction([assignment])

    def add_assignment_disjunction(self, assignments):
        self.combinatorial_assignments.append(assignments)

    def add_negative_clause(self, negative_clause):
        self.neg_clauses.append(negative_clause)

    def combine(self, other):
        """Combines two constraint systems to a new system"""
        combined = ConstraintSystem()
        combined.combinatorial_assignments = (self.combinatorial_assignments +
                                              other.combinatorial_assignments)
        combined.neg_clauses = self.neg_clauses + other.neg_clauses
        return combined

    def copy(self):
        other = ConstraintSystem()
        other.combinatorial_assignments = list(self.combinatorial_assignments)
        other.neg_clauses = list(self.neg_clauses)
        return other

    def dump(self):
        print("AssignmentSystem:")
        for comb_assignment in self.combinatorial_assignments:
            disj = " or ".join([str(assig) for assig in comb_assignment])
            print("  ASS: ", disj)
        for neg_clause in self.neg_clauses:
            print("  NEG: ", str(neg_clause))

    def is_solvable(self):
        """Check whether the combinatorial assignments include at least
           one consistent assignment under which the negative clauses
           are satisfiable"""
        for assignments in itertools.product(*self.combinatorial_assignments):
            combined = self._combine_assignments(assignments)
            if not combined.is_consistent():
                continue
            if self._all_clauses_satisfiable(combined):
                return True
        return False



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from downward import suites
import configs

import common_setup


REVS = ["issue436-base", "issue436-v1"]
LIMITS = {"search_time": 1800}
SUITE = suites.suite_satisficing_with_ipc11()

default_configs_satisficing = configs.default_configs_satisficing(extended=True)
CONFIGS = {}
for name in ['lazy_greedy_add', 'eager_greedy_ff', 'eager_greedy_add', 'lazy_greedy_ff', 'pareto_ff']:
    CONFIGS[name] = default_configs_satisficing[name]
    
exp = common_setup.IssueExperiment(
    search_revisions=REVS,
    configs=CONFIGS,
    suite=SUITE,
    limits=LIMITS,
    )

exp.add_absolute_report_step()
exp.add_comparison_table_step()


exp()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from downward import suites
from lab.reports import Attribute, gm

from common_setup import IssueConfig, IssueExperiment
from relativescatter import RelativeScatterPlotReport

def main(revisions=None):
    suite = suites.suite_optimal_with_ipc11()

    configs = {
        IssueConfig('astar-lmcut', ['--search', 'astar(lmcut())']),
    }

    exp = IssueExperiment(
        revisions=revisions,
        configs=configs,
        suite=suite,
        test_suite=['depot:pfile1'],
        processes=4,
        email='martin.wehrle@unibas.ch',
    )

    exp.add_comparison_table_step()
    
    exp.add_report(
        RelativeScatterPlotReport(
            attributes=["memory"],
            filter_config=["issue499-base-astar-lmcut", "issue499-v1-astar-lmcut"],
            get_category=lambda run1, run2: run1.get("domain"),
        ),
        outfile='issue499_base_v1_memory.png'
    )

    exp.add_report(
        RelativeScatterPlotReport(
            attributes=["total_time"],
            filter_config=["issue499-base-astar-lmcut", "issue499-v1-astar-lmcut"],
            get_category=lambda run1, run2: run1.get("domain"),
        ),
        outfile='issue499_base_v1_total_time.png'
    )

    exp.add_report(
        RelativeScatterPlotReport(
            attributes=["expansions_until_last_jump"],
            filter_config=["issue499-base-astar-lmcut", "issue499-v1-astar-lmcut"],
            get_category=lambda run1, run2: run1.get("domain"),
        ),
        outfile='issue499_base_v1_expansions_until_last_jump.png'
    )
    
    exp()

main(revisions=['issue499-base', 'issue499-v1'])



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import os

from lab.environments import LocalEnvironment, BaselSlurmEnvironment
from lab.reports import Attribute, finite_sum

import common_setup
from common_setup import IssueConfig, IssueExperiment

DIR = os.path.dirname(os.path.abspath(__file__))
BENCHMARKS_DIR = os.environ["DOWNWARD_BENCHMARKS"]
REVISIONS = ["issue732-base", "issue732-v7"]
BUILDS = ["debug32", "release32"]
CONFIGS = [
    IssueConfig(
        "lama-first-{build}".format(**locals()),
        [],
        build_options=[build],
        driver_options=["--alias", "lama-first", "--build", build])
    for build in BUILDS
]
SUITE = set(
    common_setup.DEFAULT_OPTIMAL_SUITE + common_setup.DEFAULT_SATISFICING_SUITE)
ENVIRONMENT = BaselSlurmEnvironment(
    priority=0, email="malte.helmert@unibas.ch")

if common_setup.is_test_run():
    SUITE = IssueExperiment.DEFAULT_TEST_SUITE
    ENVIRONMENT = LocalEnvironment(processes=1)

exp = IssueExperiment(
    revisions=REVISIONS,
    configs=CONFIGS,
    environment=ENVIRONMENT,
)

exp.add_suite(BENCHMARKS_DIR, SUITE)
exp.add_resource('sg_parser', 'sg-parser.py', dest='sg-parser.py')
exp.add_command('sg-parser', ['{sg_parser}'])

attributes = IssueExperiment.DEFAULT_TABLE_ATTRIBUTES + [
    Attribute("sg_construction_time", functions=[finite_sum], min_wins=True),
    Attribute("sg_peak_mem_diff", functions=[finite_sum], min_wins=True),
]

exp.add_comparison_table_step(attributes=attributes)

exp.run_steps()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from downward import suites

import common_setup

CONFIGS = {
        'astar_lmcount': [
            '--search',
            'astar(lmcount(lm_merged([lm_rhw(),lm_hm(m=1)]),admissible=true,optimal=true),mpd=true)'],
}

exp = common_setup.IssueExperiment(
    search_revisions=["issue443-base", "issue443-v2"],
    configs=CONFIGS,
    suite=suites.suite_optimal_with_ipc11(),
    )

exp.add_comparison_table_step()

exp()



#! /usr/bin/env python

from lab.parser import Parser

parser = Parser()

def check_planner_exit_reason(content, props):
    error = props.get('error')
    if error != 'none' and error != 'timeout' and error != 'out-of-memory':
        print 'error: %s' % error
        return

    out_of_time = False
    out_of_memory = False
    if error == 'timeout':
        out_of_time = True
    elif error == 'out-of-memory':
        out_of_memory = True
    props['out_of_time'] = out_of_time
    props['out_of_memory'] = out_of_memory

parser.add_function(check_planner_exit_reason)

def check_perfect_heuristic(content, props):
    plan_length = props.get('plan_length')
    expansions = props.get('expansions')
    if plan_length != None:
        perfect_heuristic = False
        if plan_length + 1 == expansions:
            perfect_heuristic = True
        props['perfect_heuristic'] = perfect_heuristic

parser.add_function(check_perfect_heuristic)

def check_proved_unsolvability(content, props):
    proved_unsolvability = False
    if props['coverage'] == 0:
        for line in content.splitlines():
            if line == 'Completely explored state space -- no solution!':
                proved_unsolvability = True
                break
    props['proved_unsolvability'] = proved_unsolvability

parser.add_function(check_proved_unsolvability)

parser.parse()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import os

from downward import suites

from lab.environments import LocalEnvironment, MaiaEnvironment

from common_setup import IssueConfig, IssueExperiment, is_test_run

REVS = ["issue551-base", "issue551-v5"]
BENCHMARKS = os.path.expanduser('~/downward-benchmarks')
SUITE = suites.suite_satisficing()

CONFIGS = [
    IssueConfig("seq-sat-lama-2011", [], driver_options=["--alias", "seq-sat-lama-2011"]),
]

ENVIRONMENT = MaiaEnvironment(
    priority=0, email="manuel.heusner@unibas.ch")

if is_test_run():
    SUITE = IssueExperiment.DEFAULT_TEST_SUITE
    ENVIRONMENT = LocalEnvironment(processes=4)

exp = IssueExperiment(
    revisions=REVS,
    configs=CONFIGS,
    environment=ENVIRONMENT
)

exp.add_suite(BENCHMARKS, SUITE)

exp.add_comparison_table_step()

exp()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from lab.experiment import Experiment
from lab.steps import Step
from downward.reports.compare import CompareConfigsReport
from common_setup import get_experiment_name, get_data_dir, get_repo_base

import os

DATADIR = os.path.join(os.path.dirname(__file__), 'data')

exp = Experiment(get_data_dir())

exp.add_fetcher(os.path.join(DATADIR, 'e2013101802-pho-seq-constraints-eval'), filter_config_nick="astar_pho_seq_no_onesafe")
exp.add_fetcher(os.path.join(DATADIR, 'issue527-v2-eval'), filter_config_nick="astar_occ_seq")

exp.add_report(CompareConfigsReport(
    [
     ('869fec6f843b-astar_pho_seq_no_onesafe', 'issue527-v2-astar_occ_seq'),
    ],
    attributes=[
                'coverage',
                'total_time',
                'expansions',
                'evaluations',
                'generated',
                'expansions_until_last_jump',
                'error',
                ],
    )
)



exp()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from downward import configs, suites

import common_setup


SEARCH_REVS = ["issue439-base", "issue439-v1"]
LIMITS = {"search_time": 1800}
SUITE = suites.suite_satisficing_with_ipc11()

configs_satisficing_core = configs.configs_satisficing_core()
CONFIGS = {}
for name in ["eager_greedy_add", "eager_greedy_ff",
             "lazy_greedy_add", "lazy_greedy_ff"]:
    CONFIGS[name] = configs_satisficing_core[name]


exp = common_setup.IssueExperiment(
    revisions=SEARCH_REVS,
    configs=CONFIGS,
    suite=SUITE,
    limits=LIMITS,
    )
exp.add_search_parser("custom-parser.py")

attributes = attributes=exp.DEFAULT_TABLE_ATTRIBUTES + ["init_time"]
exp.add_absolute_report_step(attributes=attributes)
exp.add_comparison_table_step(attributes=attributes)
exp.add_report(common_setup.RegressionReport(
    revision_nicks=exp.revision_nicks,
    config_nicks=CONFIGS.keys(),
    attributes=attributes))

exp()



from mercurial import cmdutil
from mercurial import util
try:
    # Mercurial >= 1.8
    # Due to mercurial's lazy importing we need to import the match function
    # directly to raise an ImportError if the scmutil module is not present.
    from mercurial.scmutil import match
    def match_func(repo, ctx, patterns, options):
        return match(ctx, patterns, options)
except ImportError:
    def match_func(repo, ctx, patterns, options):
        return cmdutil.match(repo, patterns, options)


cmdtable = {}
try:
    # use new registrar.command decorator to register uncrustify later
    from mercurial import registrar
    command = registrar.command(cmdtable)
except (ImportError, AttributeError):
    # make own decorator valid for the old way
    def command(name, opts, synopsis):
        def _decorator(func):
            cmdtable[name] = (func, opts, synopsis)
            return func
        return _decorator


import errno
import os
import os.path
import subprocess

SUFFIX = ".uncrustify.temp"


def _call_subprocesses(cmddesc, *cmd_lists):
    for cmd_list in cmd_lists:
        try:
            exitcode = subprocess.call(cmd_list)
        except OSError as e:
            # If command is not found, swallow error and try next command.
            if e.errno != errno.ENOENT:
                raise util.Abort("error running %s: %s" % (cmddesc, e))
        else:
            return exitcode
    raise util.Abort("could not find %s -- not installed?" % cmddesc)


def _run_uncrustify(config_file, filenames):
    """Call uncrustify, with some fancy additional behaviour:
    1. Make sure that we don't call with an empty file list since
       then uncrustify would wait for input on stdin.
    2. Don't call on empty files since uncrustify doesn't like them.
       Instead, manually create empty outputs for empty inputs.
    3. Barf if the output files for uncrustify already exist.
    """
    inputs = []
    for filename in filenames:
        temp_file = filename + SUFFIX
        if os.path.exists(temp_file):
            raise util.Abort("cannot run uncrustify: %s exists" % temp_file)
        if os.stat(filename).st_size:
            inputs.append(filename)
        else:
            # Create empty fake output file.
            with open(temp_file, "w"):
                pass
    if inputs:
        cmd = ["uncrustify", "-q", "-c", config_file,
               "--suffix", SUFFIX] + inputs
        returncode = _call_subprocesses("uncrustify", cmd)
        if returncode:
            uncrustify_version = subprocess.check_output(
                ["uncrustify", "--version"]).strip()
            raise util.Abort(
                "uncrustify exited with {returncode}. Are you using an "
                "outdated version? We require uncrustify 0.67, you "
                "have {uncrustify_version}. Please consult "
                "www.fast-downward.org/ForDevelopers/Uncrustify".format(
                **locals()))


def _run_diff(oldfile, newfile):
    """Run colordiff if it can be found, and plain diff otherwise."""
    # TODO: It may be nicer to use the internal diff engine for this.
    #       For one, this would use the correct colors set up for hg
    #       diff rather than the colors set up for colordiff. It's not
    #       clear to me how this can be done though, and if it is
    #       worth the bother.
    _call_subprocesses("diff or colordiff",
                       ["colordiff", "-u", oldfile, newfile],
                       ["diff", "-u", oldfile, newfile])


def _get_files(repo, patterns, options):
    """Return all files in the working directory that match the
    patterns and are tracked (clean, modified or added). Ignored or
    unknown files are only matched when given literally.

    If patterns is empty, match all tracked files.

    Supports options['include'] and options['exclude'] which work like
    the --include and --exclude options of hg status.
    """
    ctx = repo[None]
    match = match_func(repo, ctx, patterns, options)
    try:
        status = ctx.status(listclean=True, listignored=True, listunknown=True)
    except TypeError:
        # Compatibility with older Mercurial versions.
        status = ctx.status(clean=True, ignored=True, unknown=True)
    modified = status[0]
    added = status[1]
    unknown = status[4]
    ignored = status[5]
    clean = status[6]
    files = []
    for file_list in [clean, modified, added]:
        for filename in file_list:
            if match(filename):
                files.append(filename)
    for file_list in [ignored, unknown]:
        for filename in file_list:
            if match.exact(filename):
                files.append(filename)
    return files


@command("uncrustify",
         [("d", "diff", None,
           "show diff of changes that --modify would make"),
          ("m", "modify", None,
           "actually modify the files that need to be uncrustified"),
          ("", "no-backup", None,
           "do not save backup copies of modified files"),
          ("", "show-clean", None,
           "also list clean source files"),
          ("I", "include", [],
           "include names matching the given patterns", "PATTERN"),
          ("X", "exclude", [],
           "exclude names matching the given patterns", "PATTERN"),
          ],
         "[OPTION]... [FILE]...")
def uncrustify(ui, repo, *patterns, **options):
    """Run uncrustify on the specified files or directories.

    If no files are specified, operates on the whole working
    directory.

    Note: Files that don't have a .cc or .h suffix are always ignored,
    even if specified on the command line explicitly.

    By default, prints a list of files that are not clean according to
    uncrustify, using a similar output format as with hg status. No
    changes are made to the working directory.

    With the --diff option, prints the changes suggested by uncrustify
    in unified diff format. No changes are made to the working
    directory.

    With the --modify option, actually performs the changes suggested
    by uncrustify. The original (dirty) files are backed up with a
    .crusty suffix. Existing files with such a suffix are silently
    overwritten. To disable these backups, use --no-backup.

    This command always operates on the working directory, not on
    arbitrary repository revisions.

    Returns 0 on success.
    """
    if options["diff"] and options["modify"]:
        raise util.Abort("cannot specify --diff and --modify at the same time")

    if options["diff"]:
        mode = "diff"
    elif options["modify"]:
        mode = "modify"
    else:
        mode = "status"

    no_backup = options["no_backup"]
    show_clean = options["show_clean"]

    paths = [path for path in _get_files(repo, patterns, options)
             if path.endswith((".cc", ".h"))]


    uncrustify_cfg = repo.pathto(".uncrustify.cfg")
    relpaths = [repo.pathto(path) for path in paths]
    if not os.path.exists(uncrustify_cfg):
        raise util.Abort("could not find .uncrustify.cfg in repository root")
    _run_uncrustify(uncrustify_cfg, relpaths)

    ctx = repo[None]
    for path in paths:
        relpath = repo.pathto(path)
        uncr_path = path + SUFFIX
        uncr_relpath = relpath + SUFFIX
        have_changes = (ctx[path].data() != ctx[uncr_path].data())

        if have_changes:
            if mode == "status":
                ui.write("M %s\n" % relpath, label="status.modified")
                util.unlink(uncr_relpath)
            elif mode == "diff":
                _run_diff(relpath, uncr_relpath)
                util.unlink(uncr_relpath)
            elif mode == "modify":
                if not no_backup:
                    util.rename(relpath, relpath + ".crusty")
                util.rename(uncr_relpath, relpath)
                if not ui.quiet:
                    ui.write("%s uncrustified\n" % relpath)
        else:
            if show_clean:
                if mode == "status":
                    ui.write("C %s\n" % relpath, label="status.clean")
                elif mode == "modify":
                    ui.write("%s is clean\n" % relpath)
            util.unlink(uncr_relpath)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from common_setup import IssueConfig, IssueExperiment
from relativescatter import RelativeScatterPlotReport
import suites


configs = [
    IssueConfig(
        "cegar-landmarks-10k",
        ["--search", "astar(cegar(subtasks=[landmarks()],max_states=10000))"]),
    IssueConfig(
        "cegar-landmarks-goals-900s",
        ["--search", "astar(cegar(subtasks=[landmarks(),goals()],max_time=900))"]),
]
revisions = ["issue650-base", "issue650-v1"]

exp = IssueExperiment(
    revisions=revisions,
    configs=configs,
    suite=suites.suite_optimal_strips(),
    test_suite=["depot:pfile1"],
    email="jendrik.seipp@unibas.ch",
)

exp.add_comparison_table_step()

for attribute in ["memory", "total_time"]:
    for config in configs:
        exp.add_report(
            RelativeScatterPlotReport(
                attributes=[attribute],
                filter_config=["{}-{}".format(rev, config.nick) for rev in revisions],
                get_category=lambda run1, run2: run1.get("domain"),
            ),
            outfile="{}-{}-{}.png".format(exp.name, attribute, config.nick)
        )

exp()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function

import argparse
import re
import subprocess
import sys


STD_REGEX = r"(^|\s|\W)std::"

def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("cc_file", nargs="+")
    return parser.parse_args()


def check_std(cc_file):
    source_without_comments = subprocess.check_output(
        ["gcc", "-fpreprocessed", "-dD", "-E", cc_file])
    errors = []
    for line in source_without_comments.splitlines():
        if re.search(STD_REGEX, line):
            errors.append("Remove std:: from {}: {}".format(
                cc_file, line.strip()))
    return errors


def main():
    args = parse_args()

    errors = []
    for cc_file in args.cc_file:
        errors.extend(check_std(cc_file))
    for error in errors:
        print(error)
    if errors:
        sys.exit(1)


if __name__ == "__main__":
    main()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from downward import suites

import common_setup


REVS = ["issue392-v2"]
LIMITS = {"search_time": 300}

CONFIGS = {}
for randomize in ["false", "true"]:
    for pref_first in ["false", "true"]:
        CONFIGS["lama-nonunit-randomize-%(randomize)s-pref_first-%(pref_first)s" % locals()] = [
            "--heuristic", "hlm1,hff1=lm_ff_syn(lm_rhw(reasonable_orders=true,lm_cost_type=ONE,cost_type=ONE))",
            "--heuristic", "hlm2,hff2=lm_ff_syn(lm_rhw(reasonable_orders=true,lm_cost_type=PLUSONE,cost_type=PLUSONE))",
            "--search",
            "iterated(["
                "lazy_greedy([hff1,hlm1],preferred=[hff1,hlm1],randomize_successors=%(randomize)s,preferred_successors_first=%(pref_first)s,cost_type=ONE,reopen_closed=false),"
                "lazy_greedy([hff2,hlm2],preferred=[hff2,hlm2],randomize_successors=%(randomize)s,preferred_successors_first=%(pref_first)s,reopen_closed=false),"
                "lazy_wastar([hff2,hlm2],preferred=[hff2,hlm2],randomize_successors=%(randomize)s,preferred_successors_first=%(pref_first)s,w=5),"
                "lazy_wastar([hff2,hlm2],preferred=[hff2,hlm2],randomize_successors=%(randomize)s,preferred_successors_first=%(pref_first)s,w=3),"
                "lazy_wastar([hff2,hlm2],preferred=[hff2,hlm2],randomize_successors=%(randomize)s,preferred_successors_first=%(pref_first)s,w=2),"
                "lazy_wastar([hff2,hlm2],preferred=[hff2,hlm2],randomize_successors=%(randomize)s,preferred_successors_first=%(pref_first)s,w=1)],"
                "repeat_last=true,continue_on_fail=true)" % locals()
        ]

SUITE = sorted(set(suites.suite_satisficing_with_ipc11()) &
               set(suites.suite_diverse_costs()))


exp = common_setup.IssueExperiment(
    revisions=REVS,
    configs=CONFIGS,
    suite=SUITE,
    limits=LIMITS,
    )

exp.add_absolute_report_step()

exp()



def configs_optimal_core():
    return {
        # A*
        "astar_blind": [
            "--search",
            "astar(blind)"],
        "astar_h2": [
            "--search",
            "astar(hm(2))"],
        "astar_ipdb": [
            "--search",
            "astar(ipdb)"],
        "astar_lmcount_lm_merged_rhw_hm": [
            "--search",
            "astar(lmcount(lm_merged([lm_rhw(),lm_hm(m=1)]),admissible=true),mpd=true)"],
        "astar_lmcut": [
            "--search",
            "astar(lmcut)"],
        "astar_hmax": [
            "--search",
            "astar(hmax)"],
        "astar_merge_and_shrink_bisim": [
            "--search",
            "astar(merge_and_shrink("
            "merge_strategy=merge_linear(variable_order=reverse_level),"
            "shrink_strategy=shrink_bisimulation(max_states=200000,greedy=false,"
            "group_by_h=true)))"],
        "astar_merge_and_shrink_greedy_bisim": [
            "--search",
            "astar(merge_and_shrink("
            "merge_strategy=merge_linear(variable_order=reverse_level),"
            "shrink_strategy=shrink_bisimulation(max_states=infinity,threshold=1,"
            "greedy=true,group_by_h=false)))"],
        "astar_merge_and_shrink_dfp_bisim": [
            "--search",
            "astar(merge_and_shrink(merge_strategy=merge_dfp,"
            "shrink_strategy=shrink_bisimulation(max_states=50000,threshold=1,"
            "greedy=false,group_by_h=true)))"],
        "astar_selmax_lmcut_lmcount": [
            "--search",
            "astar(selmax([lmcut(),lmcount(lm_merged([lm_hm(m=1),lm_rhw()]),"
            "admissible=true)],training_set=1000),mpd=true)"],
    }


def configs_satisficing_core():
    return {
        # A*
        "astar_goalcount": [
            "--search",
            "astar(goalcount)"],
        # eager greedy
        "eager_greedy_ff": [
            "--heuristic",
            "h=ff()",
            "--search",
            "eager_greedy(h, preferred=h)"],
        "eager_greedy_add": [
            "--heuristic",
            "h=add()",
            "--search",
            "eager_greedy(h, preferred=h)"],
        "eager_greedy_cg": [
            "--heuristic",
            "h=cg()",
            "--search",
            "eager_greedy(h, preferred=h)"],
        "eager_greedy_cea": [
            "--heuristic",
            "h=cea()",
            "--search",
            "eager_greedy(h, preferred=h)"],
        # lazy greedy
        "lazy_greedy_ff": [
            "--heuristic",
            "h=ff()",
            "--search",
            "lazy_greedy(h, preferred=h)"],
        "lazy_greedy_add": [
            "--heuristic",
            "h=add()",
            "--search",
            "lazy_greedy(h, preferred=h)"],
        "lazy_greedy_cg": [
            "--heuristic",
            "h=cg()",
            "--search",
            "lazy_greedy(h, preferred=h)"],
    }


def configs_optimal_ipc():
    return {
        "seq_opt_merge_and_shrink": ["ipc", "seq-opt-merge-and-shrink"],
        "seq_opt_fdss_1": ["ipc", "seq-opt-fdss-1"],
        "seq_opt_fdss_2": ["ipc", "seq-opt-fdss-2"],
    }


def configs_satisficing_ipc():
    return {
        "seq_sat_lama_2011": ["ipc", "seq-sat-lama-2011"],
        "seq_sat_fdss_1": ["ipc", "seq-sat-fdss-1"],
        "seq_sat_fdss_2": ["ipc", "seq-sat-fdss-2"],
    }


def configs_optimal_extended():
    return {
        # A*
        "astar_lmcount_lm_merged_rhw_hm_no_order": [
            "--search",
            "astar(lmcount(lm_merged([lm_rhw(),lm_hm(m=1)]),admissible=true),mpd=true)"],
    }


def configs_satisficing_extended():
    return {
        # eager greedy
        "eager_greedy_alt_ff_cg": [
            "--heuristic",
            "hff=ff()",
            "--heuristic",
            "hcg=cg()",
            "--search",
            "eager_greedy(hff,hcg,preferred=[hff,hcg])"],
        "eager_greedy_ff_no_pref": [
            "--search",
            "eager_greedy(ff())"],
        # lazy greedy
        "lazy_greedy_alt_cea_cg": [
            "--heuristic",
            "hcea=cea()",
            "--heuristic",
            "hcg=cg()",
            "--search",
            "lazy_greedy(hcea,hcg,preferred=[hcea,hcg])"],
        "lazy_greedy_ff_no_pref": [
            "--search",
            "lazy_greedy(ff())"],
        "lazy_greedy_cea": [
            "--heuristic",
            "h=cea()",
            "--search",
            "lazy_greedy(h, preferred=h)"],
        # lazy wA*
        "lazy_wa3_ff": [
            "--heuristic",
            "h=ff()",
            "--search",
            "lazy_wastar(h,w=3,preferred=h)"],
        # eager wA*
        "eager_wa3_cg": [
            "--heuristic",
            "h=cg()",
            "--search",
            "eager(single(sum([g(),weight(h,3)])),preferred=h)"],
        # ehc
        "ehc_ff": [
            "--search",
            "ehc(ff())"],
        # iterated
        "iterated_wa_ff": [
            "--heuristic",
            "h=ff()",
            "--search",
            "iterated([lazy_wastar(h,w=10), lazy_wastar(h,w=5), lazy_wastar(h,w=3),"
            "lazy_wastar(h,w=2), lazy_wastar(h,w=1)])"],
        # pareto open list
        "pareto_ff": [
            "--heuristic",
            "h=ff()",
            "--search",
            "eager(pareto([sum([g(), h]), h]), reopen_closed=true, pathmax=false,"
            "f_eval=sum([g(), h]))"],
        # bucket-based open list
        "bucket_lmcut": [
            "--heuristic",
            "h=lmcut()",
            "--search",
            "eager(single_buckets(h), reopen_closed=true, pathmax=false)"],
    }


def default_configs_optimal(core=True, ipc=True, extended=False):
    configs = {}
    if core:
        configs.update(configs_optimal_core())
    if ipc:
        configs.update(configs_optimal_ipc())
    if extended:
        configs.update(configs_optimal_extended())
    return configs


def default_configs_satisficing(core=True, ipc=True, extended=False):
    configs = {}
    if core:
        configs.update(configs_satisficing_core())
    if ipc:
        configs.update(configs_satisficing_ipc())
    if extended:
        configs.update(configs_satisficing_extended())
    return configs
